#! /usr/bin/env bash

horizon_def=6
market_file_def=example_files/market.csv
pev_file_def=example_files/kalundborg/home_723/pev.13.csv
power_profile_file_def=example_files/kalundborg/home_723/powerprofile.copy.1.npp_scen.3.csv
dir_def=example_files/kalundborg/home_723
time_slot_length_def=60
dapp_weight_def=8
co2_weight_def=1
en_weight_def=1
past_length_def=10
past_distribution_def=`echo $past_length_def | awk '{sum = 0; pow_2 = 1; for (i = 1; i < $1; i++) {pow_2 /= 2; printf("%.30lf:", pow_2); sum += pow_2} printf("%.30lf", 1 - sum);}'`
time_stamp=`date +%F-%H-%M-%N`
#max_lsof=`lsof -i -P +c 100 2>&1 | head -1 | awk '{if ($1 == "COMMAND") print 100; else {gsub("\\\\(", ""); gsub("\\\\)", ""); print $NF}}'`

function usage ()
{
    echo $0 | awk -v horizon_def=${horizon_def} -v market_file_def=${market_file_def} -v dir_def=${dir_def} -v pev_file_def=${pev_file_def} -v power_profile_file_def=${power_profile_file_def} -v time_slot_length_def=${time_slot_length_def} -v dapp_weight_def=${dapp_weight_def} -v co2_weight_def=${co2_weight_def} -v en_weight_def=${en_weight_def} -v past_length_def=${past_length_def} -v past_distribution_def=${past_distribution_def} '{printf("\n\n\nUsage: %s [-l horizon] [-d dir] [-pp power_profile_file] [-pe pev_file] [-mr market_file] [-co2 coefficient_for_CO2_costs] [-en coefficient_for_energy_costs] [-dapp coefficient_for_transmission_and_distribution_costs] [-tsl time_slot_length] [-p past_length_in_days] [-pd past_distribution] [-ndmilps]\n\n\nWhere:\n\thorizon is how many hours have to be forecasted for each charge/discharge action computation (default is %s)\n\tdir is a directory containing the input files needed by EBR, i.e., battery.csv, profiles.csv, and energy.csv (default is %s)\n\tpower_profile_file is the EBR input file for power profiles output by DAPP (needed only if -dapp is greater than 0, default is %s)\n\tpev_file is the EBR input file with PEV characteristics (default is %s)\n\tmarket_file is the EBR input file with costs for energy and CO2 (default is %s)\n\tcoefficient_for_CO2_costs is the coefficient for the CO2 costs in EBR objective function (default is %s)\n\tcoefficient_for_en_costs is the coefficient for the energy costs in EBR objective function (default is %s)\n\tcoefficient_for_transmission_and_distribution_costs is the coefficient for the transmission and distribution costs in EBR objective function (default is %s)\n\ttime_slot_length is the duration (in minutes) of a time-slot, i.e., the periodicity of EBR invocations (default is %s)\n\tpast_length is the number of days in the past to be used for forecast (default is %s)\n\tpast_distribution is the discounting factor for the days in the past. Format is x1:...:xn, being n=past_length, and the summation on all xi must be 1 (default is %s)\n\tIf -ndmilps is given, MILP problems generated by EBR are kept\n\n\n", $1, horizon_def, dir_def, power_profile_file_def, pev_file_def, market_file_def, co2_weight_def, en_weight_def, dapp_weight_def, time_slot_length_def, past_length_def, past_distribution_def);}'
    exit
}

function cp_fail()
{
    if [ ! \( -f $1 \) ]
    then
	echo Unable to find required file $1, exiting
	exit
    fi
    cp $1 $2
}

function FileDoesNotExist ()
{
    if [ ! \( -f $1 \) ]
    then
	echo Error: file $1 does not exist
	exit
    fi
}

function ExecInPath ()
{
    res=`which $1`
    if [ "${res}" == "" ]
    then
	echo Error: $1 is not in PATH
	exit
    fi
}

function Experiment() 
{
    horizon=$1
    market_file=$2
    pev_file=$3
    power_profile_file=$4
    dir=$5
    time_slot_length=$6
    dapp_weight=$7
    co2_weight=$8
    en_weight=$9
    past_length=${10}
    past_distribution=${11}
    do_not_delete_milps=${12}
    EXP_ID=$time_stamp

    echo -n "Initializing..."
    if [ ! \( -d aux_files \) ]
    then
	echo Unable to find directory aux_files
	exit
    fi
    mkdir -p results
    mkdir -p results/$EXP_ID
    mkdir -p results/$EXP_ID/input
    mkdir -p results/$EXP_ID/bin
    mkdir -p results/$EXP_ID/output
    mkdir -p milps
    if [ ! \( -x src/EBR/bin/EBR.AHEAD$horizon \) -o ! \( -x src/House/bin/House \) ]
    then
	make -C src HORIZON=$horizon > src/make.log 2>&1
	if [ ! \( -x src/EBR/bin/EBR_fastexps.AHEAD$horizon \) ]
	then
	    echo Unable to compile, exiting
	    exit
	fi
	rm src/make.log
    fi
    echo " done"
    echo -n "Copying input... "
    cp src/EBR/bin/EBR_fastexps.AHEAD$horizon results/$EXP_ID/bin/EBR
    cp -r $dir/* results/$EXP_ID/input/
    cp $market_file results/$EXP_ID/input/`basename $market_file`
    cp_fail aux_files/HECH_conf.ini results/$EXP_ID/input
    EBR="results/$EXP_ID/bin/EBR"
    echo " done"
    echo -n "Running... "
    $EBR --tool EBR-CTRL --configHECH results/$EXP_ID/input/HECH_conf.ini --useHistoryFile results/$EXP_ID/input/profiles.csv --DAPPPowerProfileFile results/$EXP_ID/input/`basename $power_profile_file` --marketPricesFile results/$EXP_ID/input/`basename $market_file` --dappWeight $dapp_weight --co2Obj $co2_weight --arbObj $en_weight --loggingFile results/$EXP_ID/output/results.csv --batteryFile results/$EXP_ID/input/battery.csv --energyFile results/$EXP_ID/input/energy.csv --phevFile results/$EXP_ID/input/`basename $pev_file` --profilesFile results/$EXP_ID/input/profiles.csv --hLogFile results/$EXP_ID/output/other_log.csv --timeslotLength $time_slot_length > results/$EXP_ID/output/ebr.log 2>&1
    rm -f results/$EXP_ID/output/other_log.csv
    if [ $do_not_delete_milps -eq 1 ]
    then
	mv milps results/$EXP_ID/
    else
	rm -r milps
    fi
    echo " done"
    echo "See results/$EXP_ID/output/results.csv for the experiment result"
}

min_num_args=0
max_num_args=23

if [ $# -lt ${min_num_args} -o $# -gt ${max_num_args} ]
then
	usage
fi

horizon=${horizon_def}
market_file=${market_file_def}
pev_file=${pev_file_def}
power_profile_file=${power_profile_file_def}
dir=${dir_def}
time_slot_length=${time_slot_length_def}
dapp_weight=${dapp_weight_def}
co2_weight=${co2_weight_def}
en_weight=${en_weight_def}
past_length=${past_length_def}
past_distribution=${past_distribution_def}
do_not_delete_milps=0
args_ptr=1
while [ $# -ge ${args_ptr} ]
do
    opt=`echo "$*" | awk -v arg=${args_ptr} '{print $arg}'`
    num_opts=2
    case "${opt}" in
	"-l") horizon=`echo "$*" | awk -v arg=${args_ptr} '{print $(arg + 1) + 1}'`;;
	"-d") dir=`echo "$*" | awk -v arg=${args_ptr} '{print $(arg + 1) + 1}'`;;
	"-pp") power_profile_file=`echo "$*" | awk -v arg=${args_ptr} '{print $(arg + 1) + 1}'`;;
	"-pe") pev_file=`echo "$*" | awk -v arg=${args_ptr} '{print $(arg + 1) + 1}'`;;
	"-mr") market_file=`echo "$*" | awk -v arg=${args_ptr} '{print $(arg + 1) + 1}'`;;
	"-co2") co2_weight=`echo "$*" | awk -v arg=${args_ptr} '{print $(arg + 1) + 1}'`;;
	"-en") en_weight=`echo "$*" | awk -v arg=${args_ptr} '{print $(arg + 1) + 1}'`;;
	"-dapp") dapp_weight=`echo "$*" | awk -v arg=${args_ptr} '{print $(arg + 1) + 1}'`;;
	"-tsl") time_slot_length=`echo "$*" | awk -v arg=${args_ptr} '{print $(arg + 1) + 1}'`;;
	"-p") past_length=`echo "$*" | awk -v arg=${args_ptr} '{print $(arg + 1)}'`;;
	"-pd") past_distribution=`echo "$*" | awk -v arg=${args_ptr} '{print $(arg + 1)}'`;;
	"-ndmilps") 
			do_not_delete_milps=1
			num_opts=1;;
	"-h") usage;;
	*) 
	    echo Unrecognized option `echo "$*" | awk -v arg=${args_ptr} '{print $arg}'`
	    usage;;
    esac
    ((args_ptr+=num_opts))
done

FileDoesNotExist ${history_file}
FileDoesNotExist ${prices_file}
FileDoesNotExist ${pev_file}
if [ "`which cplex`" == "" -a "`which glpsol`" == "" ]
then
    echo "Please install either CPLEX or GLPK (glpsol)"
    exit
fi
if [ $horizon -gt 25 ]
then
    echo Parameter -l must be at most 24
    exit
fi

Experiment $horizon $market_file $pev_file $power_profile_file $dir $time_slot_length $dapp_weight $co2_weight $en_weight $past_length $past_distribution $do_not_delete_milps
