#include <EBR.h>
#include <Debug.h>
#include <CommandLineInterface.h>
#include <HECH.h>


#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <limits.h>
#include <unistd.h>
#include <time.h>


#define DEBUG "EBR"
#define MINUTES_IN_HOUR 60

static int read_from_file(EBR *_this)
{
	int minutes = (_this->curr_ts[14] - '0')*10 + (_this->curr_ts[15] - '0');

	if (!(minutes % (MINUTES_IN_HOUR/_this->timeslotsInHour) == 0))
	  return 0;
	else
	  return 1;
}

int main(int argc, char **argv)
{
	CommandLineInterface * cliPtr = NULL;

	HECH * myHECH = NULL;
#ifdef FAST_EXPS
	House * myHouse = NULL;
#endif
	EBR *this = NULL;

	int timeslot = 0;
	double energyCost = 0;
	/* double timeslotEnergyCost = 0; */
	long timeslots_num = 0;

	time_t start, end;

	/* measures from House (enum HECH.h:Measures) */
	double measurements[MEAS_NUMBER] = { 0 };

	/* charge/discharge/nop to battery and phev (enum HECH.h:Commands)  */
	double commands[CMDS_NUMBER] = { 0 };

	/* forecast of energy exchange with grid */
	double energyForecastedForThisTimeslot = 0;
	EBR_solution solution;
	double timeslotCO2Cost, timeslotdeltaplus = -1, timeslotdeltaminus = -1;

	/**
	 *	Forecast for next TIMESLOTS_AHEAD timeslots
	 *	for enum EBR.h:Forecasts
	 */
	double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD] = {{0}};
	bzero(&solution, sizeof(EBR_solution));


	/* Command line interface */
	cliPtr = EBR_CLI(argc, argv);

	/* Create this EBR object */
	this = EBR_new(cliPtr);

	/* Set-up connect to the HECH */
#ifndef FAST_EXPS
	myHECH = EBR_houseConnect(this->configHECH);
#else
	myHECH = EBR_houseConnect(this->configHECH, cliPtr, &myHouse);
#endif
	if (!myHECH) {
		Debug_out(DEBUG, "Failed to connect to the House server\n");
		exit(1);
	}
#ifndef FAST_EXPS
	Debug_perform(DEBUG, HECH_dumpInfo(myHECH, stdout));
#endif

	/* Timer_start("EBR main algorithm"); */
	time(&start);

	EBR_printBanner(stdout);

	timeslot = 1;

	while (1) {

		int answer, i;

		for (i = 0; i < CMDS_NUMBER; i++) {
			commands[i] = 0;
		}

		Debug_out(DEBUG, "Timeslot = %d\n", timeslot);


		/* Get current hour, for ToU purposes */
		if (this->tariff == TARIFF_TOU) {
			EBR_updateCurrentHour(this);
			Debug_out(DEBUG, "Current hour = %d\n", this->currentHour);
		}


		/* Read measurements from the House */
		Debug_out(DEBUG, "Read measurements from House server\n");
#ifndef FAST_EXPS
		answer = HECH_requestMeasurements(myHECH, measurements);
#else
		answer = HECH_requestMeasurements(myHECH, myHouse, measurements, read_from_file(this));
#endif
		if (answer == HECH_getControlStopSignal(myHECH)) {
			Debug_out(DEBUG, "Control STOP signal sent from House server\n");
			break;
		}
		else if (answer == -2) {
			Debug_out(DEBUG, "Cannot get measurements from House server\n");
			break;
		}
		Debug_out(DEBUG, "Measurements from House : energy from grid = %lg, consumption = %lg, production = %lg, battery SoC = %lg, PHEV SoC = %lg, PHEV ready hours = %lg\n", measurements[MEAS_ENERGY], measurements[MEAS_CONSUMPTION], measurements[MEAS_PRODUCTION], measurements[MEAS_BATTERY], measurements[MEAS_PHEV], measurements[MEAS_PHEV_READY_HOURS]);


		/* Reads current price policy, if tariff is IBR  */
		if (this->tariff == TARIFF_IBR) {
			Debug_out(DEBUG, "Reads current price policy\n");
			answer = EBR_stepPricePolicy(this);
			if (answer < 0) {
				Debug_out(DEBUG, "Cannot get current price policy\n");
				break;
			}
			Debug_out(DEBUG, "Current price policy:\n");
			Debug_perform(DEBUG, EBR_dumpPricePolicy(this));
			Debug_out(DEBUG, "Reads current market prices\n");
			answer = EBR_stepMarketPrices(this);
			if (answer < 0) {
				Debug_out(DEBUG, "Cannot get current market prices\n");
				break;
			}
			Debug_out(DEBUG, "Current market prices:\n");
			Debug_perform(DEBUG, EBR_dumpMarketPrices(this));
		}
		else if (this->tariff == TARIFF_SUBST) {
			Debug_out(DEBUG, "Reads current Substation ToU tariff\n");
			answer = EBR_stepSubstTariff(this);
			if (answer < 0) {
				Debug_out(DEBUG, "Cannot get current Substation ToU tariff\n");
				break;
			}
			Debug_out(DEBUG, "Current Substation ToU tariff:\n");
			Debug_perform(DEBUG, EBR_dumpSubstTariff(this));
		}


		if (this->tool == TOOL_CTRL || this->tool == TOOL_SUBST_CTRL) {
			/* Update PAST for consumption, production, and PHEV for forecasting  */
			EBR_pastPush(this, timeslot, measurements[MEAS_CONSUMPTION], measurements[MEAS_PRODUCTION], measurements[MEAS_PHEV]);
			Debug_out(DEBUG, "Measures added to PAST for learning on future\n");
			Debug_out(DEBUG, "** PAST RECORDED FOR CURRENT TIMESLOT **\n");
			Debug_perform(DEBUG, EBR_pastDump(this, timeslot));
		}


		/* Get real PHEV command if in simulation or if training TOOL_CTRL */
		if ( this->tool == TOOL_HISTORY ||
			(this->tool == TOOL_CTRL && EBR_isPastTraining(this)) ) {
			double energy;
			EBR_readPHEVCommand(this, commands);
			Debug_out(DEBUG, "Read command PHEV = %lg\n", commands[CMDS_PHEV]);
			energy = EBR_energyFromGrid(this, measurements, commands);
			EBR_energyPrice(this, energy, &timeslotCO2Cost);
		}

		/* Get forecast and decide, if training is finished */
		if ( this->tool == TOOL_EVAL ||
			(this->tool == TOOL_CTRL && !EBR_isPastTraining(this)) ||
			(this->tool == TOOL_SUBST_CTRL && !EBR_isPastTraining(this)) ) {
			answer = EBR_forecast(this, timeslot, forecast[FORECASTS_CONSUMPTION], forecast[FORECASTS_PRODUCTION], forecast[FORECASTS_PHEV]);
			if (answer < 0) {
				Debug_out(DEBUG, "No more future to forecast. Experiment ends here.\n");
				break;
			}
			Debug_out(DEBUG, "Forecast at timeslot %d\n", timeslot);
			Debug_perform(DEBUG, EBR_printForecasts(timeslot, forecast));

			/* Decide, when needed */
			Debug_out(DEBUG, "Decide commands for battery and PHEV, if any\n");
			answer = EBR_decideCommands(this, timeslot, forecast, measurements, commands, &solution);
			if (answer < 0) {
#if MILP_INFEAS_GREEDY_MNGMT
				double energy;
				Debug_out(DEBUG, "Cannot decide with MILP, setting commands using greedy strategy for PHEV if any and if needed\n");
				if (measurements[MEAS_PHEV] >= 0 && measurements[MEAS_PHEV_READY_HOURS] <= 1) {
				  double availBattEnergy = measurements[MEAS_BATTERY]/* Battery_getStateOfCharge(this->battery) */ -  Battery_getCapacity(this->battery) *Battery_getMinPerc(this->battery)/100.0;
				  double availEnergy = (this->maxEnergy - measurements[MEAS_CONSUMPTION] + measurements[MEAS_PRODUCTION])*(this->timeslotLength/60.0) + Battery_getDischargeFactor(this->battery) * (availBattEnergy < Battery_getDischargeSpeed(this->battery)*(this->timeslotLength/60.0)? availBattEnergy : Battery_getDischargeSpeed(this->battery)*(this->timeslotLength/60.0)) /*Battery_getStateOfCharge(this->battery)*/; 
					double neededEnergy = Battery_getCapacity(this->PHEV) - measurements[MEAS_PHEV] /*Battery_getStateOfCharge(this->PHEV)*/;
					if (neededEnergy > 0) {
						double battEnergy;
						commands[CMDS_PHEV]	= (availEnergy > neededEnergy? neededEnergy : availEnergy)/(this->timeslotLength/60.0); /* it is a power, but final energy must be checked */
						commands[CMDS_PHEV] = commands[CMDS_PHEV] > Battery_getDischargeSpeed(this->PHEV)? Battery_getDischargeSpeed(this->PHEV) : commands[CMDS_PHEV];
						battEnergy = commands[CMDS_PHEV]/Battery_getDischargeFactor(this->battery);
						commands[CMDS_BATTERY] = battEnergy*(this->timeslotLength/60.0) > availBattEnergy? availBattEnergy : battEnergy; /* same as above */
						commands[CMDS_BATTERY] = commands[CMDS_BATTERY] > Battery_getDischargeSpeed(this->battery)? -Battery_getDischargeSpeed(this->battery) : -commands[CMDS_BATTERY];
						availEnergy = measurements[MEAS_CONSUMPTION] - measurements[MEAS_PRODUCTION] + commands[CMDS_PHEV] + commands[CMDS_BATTERY];
						solution.dp = availEnergy > this->pricePolicy[PRICEPOLICY_MAX][0]? availEnergy - this->pricePolicy[PRICEPOLICY_MAX][0] : 0;
						solution.dm = availEnergy < this->pricePolicy[PRICEPOLICY_MIN][0]? this->pricePolicy[PRICEPOLICY_MIN][0] - availEnergy : 0;
					}
					else {
						commands[CMDS_BATTERY] = 0;
						commands[CMDS_PHEV] = 0;
					}

				}
				else {
					energyForecastedForThisTimeslot = 0;
					commands[CMDS_BATTERY] = 0;
					commands[CMDS_PHEV] = 0;
				}
				energy = EBR_energyFromGrid(this, measurements, commands);
				energyForecastedForThisTimeslot = energy;
				EBR_energyPrice(this, energy, &timeslotCO2Cost);
#else
				double energy;
				Debug_out(DEBUG, "Cannot decide, setting commands and energy to 0\n");
				energyForecastedForThisTimeslot = 0;
				commands[CMDS_BATTERY] = 0;
				commands[CMDS_PHEV] = 0;
				energy = EBR_energyFromGrid(this, measurements, commands);
				EBR_energyPrice(this, energy, &timeslotCO2Cost);
#endif
			}
			Debug_out(DEBUG, "Decided commands : battery = %lg, PHEV = %lg\n", commands[CMDS_BATTERY], commands[CMDS_PHEV]);
		}


		/* For Substation only, store energy in Window */
		if (this->tool == TOOL_SUBST_CTRL) {
			if (EBR_isPastTraining(this)) {
				commands[CMDS_BATTERY] = 0;
				commands[CMDS_PHEV] = 0;
				energyForecastedForThisTimeslot = EBR_energyFromGrid(this, measurements, commands);
				EBR_energyPrice(this, energyForecastedForThisTimeslot, &timeslotCO2Cost);
			}
			EBR_substEnergyWindowPush(this, energyForecastedForThisTimeslot);
			Debug_out(DEBUG, "Energy Window in the past:\n");
			Debug_perform(DEBUG, EBR_dumpEnergyWindow(this));
		}

		/* For Substation simulation */
		if (this->tool == TOOL_SUBST_HISTORY) {
			commands[CMDS_BATTERY] = 0;
			commands[CMDS_PHEV] = 0;
			energyForecastedForThisTimeslot = EBR_energyFromGrid(this, measurements, commands);
			EBR_energyPrice(this, energyForecastedForThisTimeslot, &timeslotCO2Cost);
		}


		/* Send commands to the House */
		Debug_out(DEBUG, "Send commands to House server\n");
#ifndef FAST_EXPS
		answer = HECH_sendCommands(myHECH, commands);
#else
		answer = HECH_sendCommands(myHECH, myHouse, commands);
#endif
		if (answer < 0) {
			Debug_out(DEBUG, "Commands not accepted\n");
		}
		Debug_out(DEBUG, "Sent commands to House : battery = %lg, PHEV = %lg\n", commands[CMDS_BATTERY], commands[CMDS_PHEV]);


		/* Log actual measurements with previous commands */
		/* e(t + 1) = d(t + 1) + beta_1 * a_1(t) + beta_2 * a_2(t) + r(t + 1) */
		EBR_logMeasurements(this, measurements, commands, forecast, &solution, EBR_isPastTraining(this)? 1 : (this->timeslotLength/60.0));
		Debug_out(DEBUG, "Logged measurements in %s\n", this->loggingFile);


#if 0
		/* Store total cost */
		energyCost += timeslotEnergyCost;
#endif
		if (EBR_isPastTraining(this)) {
		  this->curr_ts_time_t += SECONDS_IN_HOURS;
		}
		else {
		  this->curr_ts_time_t += this->timeslotLength*SECONDS_IN_MINUTES;
		}
		this->curr_ts_time = localtime(&this->curr_ts_time_t);
		/* if (this->curr_ts_time->tm_isdst != this->save_isdst) { */
		/*   this->curr_ts_time_t += SECONDS_IN_HOURS*(this->save_isdst); */
		/*   this->curr_ts_time = localtime(&this->curr_ts_time_t); */
		/* } */
		sprintf(this->curr_ts, "%d-%02d-%02d %02d:%02d:%02d", this->curr_ts_time->tm_year + POSIX_STARTING_YEAR, this->curr_ts_time->tm_mon + 1, this->curr_ts_time->tm_mday, this->curr_ts_time->tm_hour, this->curr_ts_time->tm_min, this->curr_ts_time->tm_sec);
		timeslots_num++;


		/* Sleep for the requested time */
//		Debug_out(DEBUG, "Sleep for one hour\n");
//		sleep(60*60);

		/* needed for forecast */
		timeslot++;
		if (timeslot > TIMESLOTS_IN_A_DAY) {
			timeslot = 1;
		}
		EBR_nextPastTraining(this);
	}

	/* Timer_end("EBR main algorithm"); */
	time(&end);
	printf("\nExecution time: %lg seconds\n", difftime(end, start));

#if 0
	printf("\n** TOTAL   ENERGY BILL      = %lg\n", energyCost);
	printf("** MONTHLY ENERGY BILL MEAN = %lg\n", energyCost/((double)timeslots_num/30.0));
	printf("** DAILY   ENERGY BILL MEAN = %lg\n\n", energyCost/(double)timeslots_num);
#endif

#ifndef FAST_EXPS
	if (myHECH != NULL) EBR_houseDisconnect(&myHECH);
#else
	if (myHECH != NULL) EBR_houseDisconnect(&myHECH, &myHouse);
#endif
	if (this != NULL) EBR_free(&this);
	/* if (cliPtr != NULL) CommandLineInterface_free(&cliPtr); */

	return 0;
}
