#include <EBR.h>
#include <Debug.h>
#include <CommandLineInterface.h>
#include <StringUtils.h>
#include <Battery.h>

#include <stdio.h>


#define DEBUG "EBR"


void EBR_printBanner(FILE *file)
{
  fprintf(file, "*** This is %s version %s\n", PACKAGE_STRING, VERSION_STRING);
  fprintf(file, "*** Copyright %s\n", COPYRIGHT_STRING);
  fflush(NULL); /* to flush all open output streams before any other output */
}


CommandLineInterface * EBR_CLI(int argc, char ** argv)
{
	CommandLineInterface * cliPtr = NULL;
	int i;

	printf("Command line:");
	for (i = 0; i < argc; i++) {
		printf(" %s", argv[i]);
	}
	printf("\n");
	fflush(NULL);

	cliPtr = CommandLineInterface_new(argv[0], "SmartHG EBR");
	/* Add CLI options */
	CommandLineInterface_addStringOption_withDefault(cliPtr, "configHECH", "HECH_conf.ini", "HECH Configuration file with info of House server");

	CommandLineInterface_addStringOption_withDefault(cliPtr, "tool", TOOL_DEFAULT, "Choose tool to execute among {EBR-CTRL, EBR-EVAL, EBR-HISTORY} with the following meaning:\n* EBR-CTRL for battery controller within a House (forecast of future based on past, PHEV commands for training period in --useHistoryFile)\n* EBR-EVAL for evaluation of optimization strategy (real future taken from file --useHistoryFile)\n* EBR-HISTORY for only simulating real data within a House (real future taken from file --useHistoryFile)");

	/* CommandLineInterface_addStringOption_withDefault(cliPtr, "tariff", TARIFF_DEFAULT, "Choose tariff to apply among {FLAT, IBR, ToU} with the following meaning:\n* FLAT for flat tariff, constant in time and power usage (buy tariff taken from arg --FLATBuy, sell tariff taken from arg --FLATSell)\n* IBR for Inclining Block Rate tariff, as a function of time and power usare (price policy taken from file --pricePolicyFile, low buy tariff taken from arg --IBRBuyLow, high buy tariff taken from --IBRBuyHigh, sell tariff taken from arg --IBRSell)\n* ToU for Time of Usage tariff, as a function of daily time and constant in power usage (tariffs taken from file --tariffsToUFile, first hour is taken from --useHistoryFile)"); */

	CommandLineInterface_addIntOption_withDefault(cliPtr, "pastLength", 10, "Number of measures to store from the past for each time-slot (only considered if passing --tool EBR-CTRL)");
	CommandLineInterface_addStringOption_withDefault(cliPtr, "pastDistribution", PAST_DISTRIBUTION, "Distribution for past usage, must be --pastLength long and must sum to 1, format \"1st:2nd:3rd:...:--pastLength-th\" (only considered if passing --tool EBR-CTRL)");

	CommandLineInterface_addStringOption_withDefault(cliPtr, "useHistoryFile", "profiles.csv", "In case of evaluation of optimization strategy, read data from this file, in CSV format (only considered if passing --tool EBR-EVAL or --tool EBR-HISTORY; for other options it is only for initial training period)");

	CommandLineInterface_addStringOption_withDefault(cliPtr, "tariffsToUFile", "tariffs_house_ToU.csv", "File with ToU tariffs for daily hours, in CSV format (only considered if passing --tariff ToU)");
	CommandLineInterface_addStringOption_withDefault(cliPtr, "DAPPPowerProfileFile", "dapp_output_power_profile.csv", "File with power profile output by DAPP, in CSV format");
	CommandLineInterface_addStringOption_withDefault(cliPtr, "marketPricesFile", "market.csv", "File with prices from energy market, prices for regulation and CO2 conversions, in CSV format (only considered if passing --tariff IBR; prices for regulation not considered)");
	/* CommandLineInterface_addDoubleOption_withDefault(cliPtr, "taxConversion", 0.12/2.87, "Value to eliminate taxes from prices"); */
	CommandLineInterface_addDoubleOption_withDefault(cliPtr, "dappWeight", 0.0, "Times the distribution costs more if user goes out of price policy");
	CommandLineInterface_addDoubleOption_withDefault(cliPtr, "co2Obj", 1.0, "Coefficient for CO2 part in objective function"); /* indeed, should always be 1 */
	CommandLineInterface_addDoubleOption_withDefault(cliPtr, "arbObj", 1.0, "Coefficient for arbitrage part in objective function");
	CommandLineInterface_addDoubleOption_withDefault(cliPtr, "TDPrice", 0.12, "Cost of transmission and distribution (EUR/kWh)");
	/* CommandLineInterface_addDoubleOption_withDefault(cliPtr, "IBRBuyLow", 0.1, "IBR low tariff for BUYing energy from the Grid (only considered if passing --tariff IBR)"); */
	/* CommandLineInterface_addDoubleOption_withDefault(cliPtr, "IBRBuyHigh", 0.3, "IBR high tariff for BUYing energy from the Grid (only considered if passing --tariff IBR)"); */
	/* CommandLineInterface_addDoubleOption_withDefault(cliPtr, "IBRSell", 0.05, "IBR tariff for SELLing energy to the Grid (only considered if passing --tariff IBR)"); */

	/* CommandLineInterface_addDoubleOption_withDefault(cliPtr, "FLATBuy", 0.2, "FLAT tariff for BUYing energy from the Grid (only considered if passing --tariff FLAT)"); */
	/* CommandLineInterface_addDoubleOption_withDefault(cliPtr, "FLATSell", 0.05, "FLAT tariff for SELLing energy to the Grid (only considered if passing --tariff FLAT)"); */

	CommandLineInterface_addStringOption_withDefault(cliPtr, "loggingFile", "EBR_log.csv", "Output file name for EBR logging file, in CSV format");
	CommandLineInterface_addStringOption_withDefault(cliPtr, "batteryFile", "battery.csv", "File with properties of battery, in CSV format");
	CommandLineInterface_addStringOption_withDefault(cliPtr, "energyFile", "energy.csv", "File with contractual information on min and max energy from/to the Grid, in CSV format");

	// PHEV can be NULL, in which case there is no PHEV (e.g., for --tool DAPP-S-CTRL)
	CommandLineInterface_addStringOption_withDefault(cliPtr, "phevFile", NULL, "File with properties of PHEV, in CSV format");

	CommandLineInterface_addIntOption_withDefault(cliPtr, "timeslotLength", 60, "EBR will decide an action every timeslotLength minutes (must be divisor of 60)");
	CommandLineInterface_addIntOption_withDefault(cliPtr, "pricesPerHour", 1, "Input prices change this number of times per hour");
//	CommandLineInterface_addDoubleOption_withDefault(cliPtr, "forecastHours", 24.0, "*** NOT WORKING NOW *** Number of hours ahead to consider for forecasting in optimization control strategy");
#ifdef FAST_EXPS
	/* CommandLineInterface_addStringOption_withDefault(cliPtr, "configFile", "House_conf.ini", "House Configuration file"); */
	CommandLineInterface_addStringOption_withDefault(cliPtr, "profilesFile", "profiles.csv", "Profiles file name, in CSV format");
	CommandLineInterface_addStringOption_withDefault(cliPtr, "hLogFile", "LabVIEWloggedProfiles.csv", "Output file name for real profiles logging file, in CSV format");

	/* CommandLineInterface_addFlag(cliPtr, "isSubstation", "Act as server in a Substation instead of a House (this implies that there is no PHEV, so no --phevFile must be passed)"); */
#endif

	/* CommandLineInterface_addFlag(cliPtr, "infESSCapacity", "Also minimise battery cost"); */
	/* CommandLineInterface_addDoubleOption_withDefault(cliPtr, "ESSCapacityCost", 0.0, "Cost of ESS capacity in EUR/kWh"); */
	/* CommandLineInterface_addDoubleOption_withDefault(cliPtr, "ESSPowerRateCost", 0.0, "Cost of ESS power rate in EUR/kW"); */

	/* Parse CLI options */
	CommandLineInterface_processArgs(cliPtr, argc, argv);
	return cliPtr;
}


void EBR_logMeasurements(EBR *_this, double * measurements, double * commands, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], EBR_solution *solution, double forEnergy)
{
  Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
  double energyFromGrid = EBR_energyFromGrid(_this, measurements, commands);
#if 0
  double appliedPrice = energyFromGrid != 0? energyPrice/energyFromGrid : 0;
  if (_this->tool == TOOL_SUBST_CTRL || _this->tool == TOOL_SUBST_HISTORY) {
    fprintf(_this->loggingFile_fd, "%lg,%lg,%lg,%lg,%lg,%lg,%lg",
	    energyFromGrid,
	    energyForecast,
	    measurements[MEAS_CONSUMPTION],
	    measurements[MEAS_BATTERY],
	    commands[CMDS_BATTERY],
	    energyPrice,
	    appliedPrice);
  }
  else {
#endif
    fprintf(_this->loggingFile_fd, "%s,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg",
	    _this->curr_ts,
	    energyFromGrid*forEnergy, energyFromGrid,
	    solution->e*(_this->timeslotLength/60.0), solution->e, 
	    measurements[MEAS_CONSUMPTION] - measurements[MEAS_PRODUCTION],
	    measurements[MEAS_CONSUMPTION],
	    measurements[MEAS_PRODUCTION],
	    measurements[MEAS_BATTERY],
	    commands[CMDS_BATTERY],
	    measurements[MEAS_PHEV],
	    commands[CMDS_PHEV],
	    measurements[MEAS_PHEV] + (commands[CMDS_PHEV] > 0? Battery_getChargeFactor(_this->PHEV) : 1.0)*commands[CMDS_PHEV]*_this->timeslotLength/60.0,
	    measurements[MEAS_PHEV_READY_HOURS],
	    _this->co2Conversion[0],
	    _this->energyMarket[0],
	    solution->tilde_e*(_this->timeslotLength/60.0),
	    solution->dp,
	    solution->dm,
	    _this->pricePolicy[PRICEPOLICY_MIN][0],
	    _this->pricePolicy[PRICEPOLICY_MAX][0]
	    );
#if 0
  }
#endif
  /* if price policy dump also max price policy  */
  if (_this->tariff == TARIFF_IBR) {
    fprintf(_this->loggingFile_fd, ",%lg", _this->pricePolicy[PRICEPOLICY_MAX][0]);
  }
  /* if there is a forecast based on heuristic, dump all forecast */
  if (_this->tool == TOOL_CTRL) {
    int i;
    for (i = 0; i < TIMESLOTS_AHEAD-1; i++) {
      fprintf(_this->loggingFile_fd, ",%lg,%lg,%lg",
	      forecast[FORECASTS_CONSUMPTION][i],
	      forecast[FORECASTS_PRODUCTION][i],
	      forecast[FORECASTS_PHEV][i]);
    }
  }
  else if (_this->tool == TOOL_SUBST_CTRL) {
    int i;
    for (i = 0; i < TIMESLOTS_AHEAD-1; i++) {
      fprintf(_this->loggingFile_fd, ",%lg", forecast[FORECASTS_CONSUMPTION][i]);
    }
  }
  fprintf(_this->loggingFile_fd, "\n");
  fflush(_this->loggingFile_fd);
}


void EBR_printForecasts(int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD])
{
	int i, j;
	printf("*** FORECAST timeslot = %d\n", timeslot);
	for (i = 0; i < FORECASTS_NUMBER; i++) {
		printf("forecast[%d]: ", i);
		for (j = 0; j < TIMESLOTS_AHEAD; j++) {
			printf("%lg ", forecast[i][j]);
		}
		printf("\n");
	}
	fflush(NULL);

}


void EBR_updateCurrentHour(EBR *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, _this->tariff == TARIFF_TOU, "_this->tariff != TARIFF_TOU\n");

	static char line[MAX_BUF_LEN];

	if (_this->currentHour == -1) {
		/* Read from pricepolicyfile */
		char *token = NULL;
		FILE * history_fd;

		history_fd = fopen(_this->useHistoryFile, "r");
		if (!history_fd) {
			perror("Errors opening useHistoryFile for understanding first hour (ToU)");
			exit(1);
		}

		/* Discard first line */
		if (!fgets(line, MAX_BUF_LEN, history_fd)) {
			fprintf(stderr, "Problems reading from %s\n", _this->useHistoryFile);
			exit(1);
		}
		if (!fgets(line, MAX_BUF_LEN, history_fd)) {
			fprintf(stderr, "Problems reading from %s\n", _this->useHistoryFile);
			exit(1);
		}
		if (line[strlen(line)-1] == '\n') {
			line[strlen(line)-1] = '\0';
		}
		Debug_out(DEBUG, "History file line for understanding first hour (ToU): %s\n", line);

		/* tokenize and put pricepolicies into pp */
		/* starting */
		token = strtok(line, ",");
		if (!token) {
			perror("ERROR in tokenize");
			exit(1);
		}
		sprintf(line, "%s", token);
		token = strtok(line, " "); /* e.g. 2014-07-04 */
		token = strtok(NULL, " "); /* e.g. 00:00:00 */
		if (!token) {
			perror("ERROR in tokenize");
			exit(1);
		}
		sprintf(line, "%s", token);
		token = strtok(line, " "); /* e.g. 00 */
		if (!token) {
			perror("ERROR in tokenize");
			exit(1);
		}
		if (StringUtils_toInt(token, &(_this->currentHour)) != 0) {
			Debug_out(DEBUG, "Problems reading first hour, value %s not accepted\n", token);
			exit(1);
		}

		fclose(history_fd);
	}
	else {
		_this->currentHour = _this->currentHour < 23? _this->currentHour + 1 : 0;
	}
}


/**
 *	measurements refers to timeslot t, commands refers to timeslot t-1
 */
double EBR_energyFromGrid(EBR *_this, double * measurements, double * commands)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	double battFactor = 0;
	double phevFactor = 0;
	double res = 0;
	if (_this->battery) {
		battFactor = commands[CMDS_BATTERY] > 0? 1 : Battery_getDischargeFactor(_this->battery);
	}
	if (_this->PHEV) {
		phevFactor = commands[CMDS_PHEV] > 0? 1 : Battery_getDischargeFactor(_this->PHEV);
	}
	Debug_out(DEBUG, "EBR_energyFromGrid: consumption = %lg\n", measurements[MEAS_CONSUMPTION]);
	Debug_out(DEBUG, "EBR_energyFromGrid: production = %lg\n", measurements[MEAS_PRODUCTION]);
	Debug_out(DEBUG, "EBR_energyFromGrid: battFactor = %lg\n", battFactor);
	Debug_out(DEBUG, "EBR_energyFromGrid: commands[CMDS_BATTERY] = %lg\n", commands[CMDS_BATTERY]);
	Debug_out(DEBUG, "EBR_energyFromGrid: commands[CMDS_PHEV] = %lg\n", commands[CMDS_PHEV]);
	res = measurements[MEAS_CONSUMPTION] + battFactor*commands[CMDS_BATTERY] + phevFactor*commands[CMDS_PHEV] - measurements[MEAS_PRODUCTION];
	/* res *= _this->timeslotLength/60.0; */
	Debug_out(DEBUG, "EBR_energyFromGrid: res = %lg\n", res);
	return res;
}



void EBR_dumpEnergyWindow(EBR *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, _this->tool == TOOL_SUBST_CTRL, "_this->tool == TOOL_SUBST_CTRL\n");

	int i;
	for (i = 0; i <= _this->energyWindowSubst_index && i < _this->energyWindowSubst_l - 1; i++) {
		printf("\t%lg", _this->energyWindowSubst[i]);
	}
	printf("\n");
	fflush(NULL);
}

