#ifndef __HOUSE_H__
#define __HOUSE_H__

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>

#include <Properties.h>
#include <Battery.h>



/* Linux defines PATH_MAX in a non standard library, so we redefine it! */
#define PATH_MAX        4096


#define MAX_BUF_LEN (1 << 20)


typedef enum meas {
	MEAS_ENERGY = 0,
	MEAS_CONSUMPTION,
	MEAS_PRODUCTION,
	MEAS_BATTERY,
	MEAS_PHEV,
	MEAS_PHEV_READY_HOURS,
	MEAS_NUMBER
} Measures;


typedef enum commands {
	CMDS_BATTERY = 0,
	CMDS_PHEV,
	CMDS_NUMBER
} Commands;



/**
 *	Information on Server
 */

typedef struct Server {
	const char * address;
	int port;

	const char * name;
	int maxMsgSize;
	int maxAcceptedConnAtTime;
	int controlStopServer;
} Server;

Server * Server_new(const char *name, const char * address, int port, int controlStopServer, int maxMsgSize, int maxAccConnAtTime);

void Server_free(Server **_this);

/** End Server **/



/**
 *  Object simulating House (LabVIEW in IMDEA labs)
 *	House is a server with two channels
 *	1. From/To EBR for measurements requests
 *	2. From/To EBR for commands to battery/PHEV
 */


typedef struct House {

	int isSubstation;

	int consumerMeasurementsPort;
	Server *measurementsConsumer;
	int consumerMeasurementsPID;

	int consumerCommandsPort;
	Server *commandsConsumer;
	int consumerCommandsPID;

	char * profilesFile;
	FILE * profilesFile_fd;
	unsigned long profilesFileCurrentLine;

	char * loggingFile;
	FILE * loggingFile_fd;

	Battery *battery;
	double batteryChargeCmd;

	Battery *PHEV;
	double phevChargeCmd;

	int timeSlotLength; // in minutes

	int pipeMeas[2]; // [0] read, [1] write
	int pipeCmds[2]; // [0] read, [1] write

	int onlySimulateProfiles; // only simulates real data from profilesFile

} House;



House * House_new(Properties * conf, const char *profilesFiles, const char *loggingFile, const char * batteryFile, const char * phevFile, int onlySimulateProfiles, int isSubstation);

/**
 *	Starts servers 
 *	i) for Consumer at consumerMeasurementsPort for measurements requests and
 *	ii)  at consumerCommandsPort for commands
 */

void House_startServers(House *_this);

pid_t House_startMeasurements(House *_this, FILE *stream);

pid_t House_startCommands(House *_this, FILE *stream);


/**
 *	Closes all servers, if need be
 */

void House_free(House ** _this);


/**
 *	Waits until all servers got control signal STOP
 */

void House_waitServers(House *);





/**
 *  Functionalities for communication on two channels
 *
 *	(1) giving measurements when requested in
 *		1) consumption
 *		2) production
 *		3) battery state of charge
 *		4) PHEV state of charge (negative if not present)
 *	Example:
 *	(1)-> %d
 *	(1)<- %d %lg %lg %lg %lg
 *
 *	(2) accepting commands
 *		1) for battery
 *		2) accepting commands for PHEV, if present
 *	Example:
 *	(2)-> %d %lg %lg
 */

long int House_measurementsReqsFromConsumer(House *_this, long int controlsignal, double * measurements);

long int House_commandsFromConsumer(House *_this, long int controlsignal, double * commands);


/**
 *	Logging functionalities
 */

#endif
