#define _BSD_SOURCE

#include <House.h>
#include <Properties.h>
#include <Debug.h>
#include <StringUtils.h>
#include <Battery.h>


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
/* pipes */
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/wait.h>


#define DEBUG "House Functions"


#define MSG_PIPE_LEN 1024




House * House_new(Properties *conf, const char *profilesFile, const char *loggingFile, const char * batteryFile, const char * phevFile, int onlySimulateProfiles, int isSubstation)
{
	House * new = (House *) calloc(1, sizeof(House));
	int controlSignalStop;
	static char line[MAX_BUF_LEN];

	if (Properties_getInt(conf, "controlSignalStop", &controlSignalStop) < 0) {
		fprintf(stderr, "\nAdd in configuration file a line\n\nint controlSignalStop = anInteger\n");
		exit(1);
	}

	if (Properties_getInt(conf, "timeSlotLength", &(new->timeSlotLength)) < 0) {
		fprintf(stderr, "\nAdd in configuration file a line\n\nint timeSlotLength = anIntegerInMinutes\n");
		exit(1);
	}

	if (Properties_getInt(conf, "CONSUMER_LISTEN_MEAS_PORT", &(new->consumerMeasurementsPort)) < 0) {
		fprintf(stderr, "\nAdd in configuration file a line\n\nint CONSUMER_LISTEN_MEAS_PORT = anInteger\n");
		exit(1);
	}
	new->measurementsConsumer = Server_new("MEASUREMENTS TO CONSUMER", NULL, new->consumerMeasurementsPort, controlSignalStop, 1024, 10);

	if (Properties_getInt(conf, "CONSUMER_LISTEN_CMDS_PORT", &(new->consumerCommandsPort)) < 0) {
		fprintf(stderr, "\nAdd in configuration file a line\n\nint CONSUMER_LISTEN_MEAS_PORT = anInteger\n");
		exit(1);
	}
	new->commandsConsumer = Server_new("COMMANDS FROM CONSUMER", NULL, new->consumerCommandsPort, controlSignalStop, 1024, 10);


	new->onlySimulateProfiles = onlySimulateProfiles;

	new->isSubstation = isSubstation;


	/** Open profiles input files and logging output file */
	new->profilesFile_fd = fopen(profilesFile, "r");
	if (!new->profilesFile_fd) {
		fprintf(stderr, "Error opening file %s\n", profilesFile);
		exit(1);
	}
	if (!fgets(line, MAX_BUF_LEN, new->profilesFile_fd)) {
		fprintf(stderr, "Problems reading from %s\n", new->profilesFile);
		exit(1);
	}
	if (line[strlen(line)-1] == '\n') {
		line[strlen(line)-1] = '\0';
	}
	Debug_out(DEBUG, "Profiles first line : %s", line);

	new->loggingFile_fd = fopen(loggingFile, "w");
	if (!new->loggingFile_fd) {
		fprintf(stderr, "Error creating file %s\n", loggingFile);
		exit(1);
	}
	if (new->isSubstation || !phevFile) {
		fprintf(new->loggingFile_fd, "control,starting,ending,energy_grid,demand,battery,battery_charge\n");
	}
	else {
		fprintf(new->loggingFile_fd, "control,starting,ending,energy_grid,consumption,production,battery,phev,battery_charge,phev_charge,phev_ready_hours\n");
	}
	fflush(new->loggingFile_fd);

	/* read battery and PHEV properties, initially set to half their capacity */
	if (!new->onlySimulateProfiles) {
		new->battery = Battery_new(batteryFile);
		Debug_out(DEBUG, "Battery:\n");
		Battery_setStateOfCharge(new->battery, Battery_getCapacity(new->battery)/2);
		Debug_perform(DEBUG, Battery_dump(new->battery, stdout));
	}
	else {
		new->battery = NULL;
		Debug_out(DEBUG, "Discarding Battery (only simulating historical data)\n");
	}

	if (!new->isSubstation && phevFile) {
		new->PHEV = Battery_new(phevFile);
		Debug_out(DEBUG, "PHEV:\n");
		Battery_setStateOfCharge(new->PHEV, -1);
		Debug_perform(DEBUG, Battery_dump(new->PHEV, stdout));
	}
	else {
		new->PHEV = NULL;
		Debug_out(DEBUG, "PHEV is NULL\n");
	}

	/* open pipes */
	if (pipe(new->pipeMeas) < 0) {
		perror("Pipe for measurements creation failed");
		exit(1);
	}
	if (pipe(new->pipeCmds) < 0) {
		perror("Pipe for commands creation failed");
		exit(1);
	}
	Debug_out(DEBUG, "Pipes for measurements and commands created\n");

	new->batteryChargeCmd = 0;
	new->phevChargeCmd = 0;

	return new;
}



void House_free(House ** _this)
{
	Debug_assert(DEBUG, *_this != NULL, "*_this == NULL\n");
	fclose((*_this)->profilesFile_fd);
	fclose((*_this)->loggingFile_fd);
	Server_free(&(*_this)->measurementsConsumer);
	Server_free(&(*_this)->commandsConsumer);
	if ((*_this)->battery) {
		Battery_free(&(*_this)->battery);
	}
	if ((*_this)->PHEV) {
		Battery_free(&(*_this)->PHEV);
	}
	free(*_this);
	*_this = NULL;
}





void House_startServers(House *_this)
{
	Debug_out(DEBUG, "Starting server for measurements reqs from Consumers\n");
	_this->consumerMeasurementsPID = House_startMeasurements(_this, stdout);
	Debug_out(DEBUG, "Server for measurements reqs from Consumers PID = %d\n", _this->consumerMeasurementsPID);

	Debug_out(DEBUG, "Starting server for commands from Consumers\n");
	_this->consumerCommandsPID = House_startCommands(_this, stdout);
	Debug_out(DEBUG, "Server for commands from Consumers PID = %d\n", _this->consumerCommandsPID);
}



void House_waitServers(House *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");

	/* Waits for all children to terminate */
	while(wait(NULL)>0);

}



/**
 *	- If there are no more data to be applied, 
		the House server (LabVIEW) sends
 (1)<- -1

 - If there was an error that obligates to stop the experiment
 (1)<- -2
 */
long int House_measurementsReqsFromConsumer(House *this, long int control, double * measurements)
{
	Debug_assert(DEBUG, this != NULL, "this == NULL\n");

	static char line[MAX_BUF_LEN];

	char *lineTmp;
	ssize_t nbytesread;
	char * token;
	char * starting;
	char * ending;
	double energyGrid = 0;
	double phev = -1;
	double prevPHEV = 0;
	static char msgPipe[MSG_PIPE_LEN];

	Debug_out(DEBUG, "Received control '%ld'\n", control);
	if (control != (this->profilesFileCurrentLine) + 1) {
		Debug_out(DEBUG, "Requesting control '%ld' while last requested control was '%ld'\n", control, this->profilesFileCurrentLine);
		return -2;
	}
	Debug_out(DEBUG, "Requested control '%ld'\n", control);

	this->batteryChargeCmd = 0;
	this->phevChargeCmd = 0;

	/* Read commands from pipe */
	/* read actual control signal from other server */
	close(this->pipeMeas[0]); // close reading from pipeMeas
	close(this->pipeCmds[1]); // close writing in pipeCmds
	if (this->profilesFileCurrentLine > 0) {
		/* first time there is no message in pipe */
		bzero(msgPipe, MSG_PIPE_LEN);
		nbytesread = read(this->pipeCmds[0], msgPipe, MSG_PIPE_LEN);
		if (nbytesread <= 0) {
			perror("Problems reading from House named pipe");
			return -2;
		}
		token = strtok(msgPipe, " ");
		if (!token) {
			Debug_out(DEBUG, "Problems tokenizing %s\n", msgPipe);
			return -2;
		}
		if (StringUtils_toDouble(token, &(this->batteryChargeCmd)) < 0) {
			Debug_out(DEBUG, "Value '%s' not accepted\n", token);
			return -2;
		}
		token = strtok(NULL, " ");
		if (!token) {
			Debug_out(DEBUG, "Problems tokenizing %s\n", msgPipe);
			return -2;
		}
		if (StringUtils_toDouble(token, &(this->phevChargeCmd)) < 0) {
			Debug_out(DEBUG, "Value '%s' not accepted\n", token);
			return -2;
		}
		Debug_out(DEBUG, "[MEASUREMENTS] FROM PIPE [Cmds] : %lg %lg\n", this->batteryChargeCmd, this->phevChargeCmd);
	}

	/* Increase control signal */
	this->profilesFileCurrentLine = (this->profilesFileCurrentLine) + 1;

	/* Write actual control signal to named pipe */
	bzero(msgPipe, MSG_PIPE_LEN);
	sprintf(msgPipe, "%ld\n", this->profilesFileCurrentLine);
	nbytesread = write(this->pipeMeas[1], msgPipe, MSG_PIPE_LEN);
	if (nbytesread <= 0) {
		perror("Problems in writing control signal on House named pipe");
		return -2;
	}
	Debug_out(DEBUG, "[MEASUREMENTS] TO PIPE [Meas] : %ld\n", this->profilesFileCurrentLine);


	/* Read current line from file and set measurements accordingly */
	if (!fgets(line, MAX_BUF_LEN, this->profilesFile_fd)) {
		Debug_out(DEBUG, "No more lines to read from profiles file %s\n", this->profilesFile);
		return this->measurementsConsumer->controlStopServer;
	}
	if (line[strlen(line)-1] == '\n') {
		line[strlen(line)-1] = '\0';
	}
	Debug_out(DEBUG, "Retrieved profiles : %s\n", line);

	/* starting */
	lineTmp = StringUtils_clone(line);
	token = strsep(&lineTmp, ",");
	if (!token) {
		perror("Problems tokenizing file line");
		return -2;
	}
	Debug_out(DEBUG, "token = '%s'\n", token);
	starting = StringUtils_clone(token);
	/* ending */
	token = strsep(&lineTmp, ",");
	if (!token) {
		perror("Problems tokenizing file line");
		return -2;
	}
	Debug_out(DEBUG, "token = '%s'\n", token);
	ending = StringUtils_clone(token);
	/* consumption */
	token = strsep(&lineTmp, ",");
	if (!token) {
		perror("Problems tokenizing file line");
		return -2;
	}
	Debug_out(DEBUG, "token = '%s'\n", token);
	if (StringUtils_toDouble(token, &measurements[MEAS_CONSUMPTION]) != 0) {
		Debug_out(DEBUG, "Problems reading consumption, value '%s' not accepted\n", token);
		return -2;
	}
	if (!this->isSubstation) {
		/* production */
		token = strsep(&lineTmp, ",");
		if (!token) {
			perror("Problems tokenizing file line");
			return -2;
		}
		Debug_out(DEBUG, "token = '%s'\n", token);
		if (StringUtils_toDouble(token, &measurements[MEAS_PRODUCTION]) != 0) {
			Debug_out(DEBUG, "Problems reading production, value '%s' not accepted\n", token);
			return -2;
		}
		/* check phev presence */
		if (this->PHEV) {
			prevPHEV = Battery_getStateOfCharge(this->PHEV);
			token = strsep(&lineTmp, ",");
			if (!token) {
				perror("Problems tokenizing file line");
				return -2;
			}
			Debug_out(DEBUG, "token = '%s'\n", token);
			if (!StringUtils_equals(token, "")) {
				if (StringUtils_toDouble(token, &phev) != 0) {
					Debug_out(DEBUG, "Problems reading phev, value %s not accepted\n", token);
					return -2;
				}
				Debug_out(DEBUG, "phev = '%lg'\n", phev);
				/* if not present before, register soc of plugging PHEV */
				/* always set if in simulation */
				if (this->onlySimulateProfiles) {
					Battery_setStateOfCharge(this->PHEV, phev);
				}
				else if (phev != -1 && Battery_getStateOfCharge(this->PHEV) == -1) {
					Battery_setStateOfCharge(this->PHEV, phev);
				}
			}
			else {
				Debug_out(DEBUG, "PHEV not present\n");
				Battery_setStateOfCharge(this->PHEV, -1);
			}
			/* phev charging kwh */
			token = strsep(&lineTmp, ",");
			if (!token) {
				perror("Problems tokenizing file line");
				return -2;
			}
			Debug_out(DEBUG, "token = '%s'\n", token);
			/* phev user desiderata: hours in which it must be ready */
			token = strsep(&lineTmp, ",");
			if (!token) {
				perror("Problems tokenizing file line");
				return -2;
			}
			Debug_out(DEBUG, "token = '%s'\n", token);
			if (StringUtils_toDouble(token, &measurements[MEAS_PHEV_READY_HOURS]) != 0) {
				Debug_out(DEBUG, "Problems reading production, value '%s' not accepted\n", token);
				return -2;
			}
		}
	}
	else { /* in case of Substation, just to be sure, zero every non related meas */
		measurements[MEAS_PRODUCTION] = 0;
		measurements[MEAS_PHEV] = 0;
		measurements[MEAS_PHEV_READY_HOURS] = 0;
	}
	free(lineTmp);

	if (this->onlySimulateProfiles) {
		this->batteryChargeCmd = 0;
		this->phevChargeCmd = 0;
		if (!this->isSubstation && this->PHEV && Battery_getStateOfCharge(this->PHEV) > -1 && prevPHEV >= 0) {
			this->phevChargeCmd = (Battery_getStateOfCharge(this->PHEV) - prevPHEV)/Battery_getChargeFactor(this->PHEV);
			if (this->phevChargeCmd < 0)
				this->phevChargeCmd = 0;
		}
		energyGrid = measurements[MEAS_CONSUMPTION] + this->phevChargeCmd - measurements[MEAS_PRODUCTION];
	}
	else {
		double factor = this->battery && this->batteryChargeCmd < 0? Battery_getDischargeFactor(this->battery) : 1;
		/* Charge or discharge battery as commanded, only if not in simulation */
		if (this->battery && this->batteryChargeCmd != 0) {
			Battery_actuate(this->battery, this->batteryChargeCmd, (double)this->timeSlotLength/60);
		}
		if (!this->isSubstation && this->PHEV && this->phevChargeCmd > 0 && Battery_getStateOfCharge(this->PHEV) > -1) {
			Battery_actuate(this->PHEV, this->phevChargeCmd, (double)this->timeSlotLength/60);
		}
		energyGrid = measurements[MEAS_CONSUMPTION] + factor * this->batteryChargeCmd + this->phevChargeCmd - measurements[MEAS_PRODUCTION];
	}

	/* Set remaining measurements fields */
	measurements[MEAS_ENERGY] = energyGrid;
	if (!this->onlySimulateProfiles && this->battery) {
		measurements[MEAS_BATTERY] = Battery_getStateOfCharge(this->battery);
	}
	else {
		measurements[MEAS_BATTERY] = 0;
	}
	if (!this->isSubstation && this->PHEV) {
		measurements[MEAS_PHEV] = Battery_getStateOfCharge(this->PHEV);
	}

	/* dump loggingFile in CSV */
	if (this->isSubstation || !this->PHEV) {
		fprintf(this->loggingFile_fd, "%lu,%s,%s,%lg,%lg,%lg,%lg\n", this->profilesFileCurrentLine, starting, ending, measurements[MEAS_ENERGY], measurements[MEAS_CONSUMPTION], measurements[MEAS_BATTERY], this->batteryChargeCmd);
	}
	else {
		if (Battery_getStateOfCharge(this->PHEV) == -1) {
			fprintf(this->loggingFile_fd, "%lu,%s,%s,%lg,%lg,%lg,%lg,,%lg,,\n", this->profilesFileCurrentLine, starting, ending, measurements[MEAS_ENERGY], measurements[MEAS_CONSUMPTION], measurements[MEAS_PRODUCTION], measurements[MEAS_BATTERY], this->batteryChargeCmd);
		}
		else {
			fprintf(this->loggingFile_fd, "%lu,%s,%s,%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg\n", this->profilesFileCurrentLine, starting, ending, measurements[MEAS_ENERGY], measurements[MEAS_CONSUMPTION], measurements[MEAS_PRODUCTION], measurements[MEAS_BATTERY], measurements[MEAS_PHEV], this->batteryChargeCmd, this->phevChargeCmd, measurements[MEAS_PHEV_READY_HOURS]);
		}
	}
	fflush(this->loggingFile_fd);

	return this->profilesFileCurrentLine;
}


long int House_commandsFromConsumer(House *this, long int control, double * commands)
{
	Debug_assert(DEBUG, this != NULL, "this == NULL\n");
	static char msgPipe[MSG_PIPE_LEN];
	ssize_t nbytesread;

	/* read actual control signal from other server */
	close(this->pipeMeas[1]); // close writing from pipeMeas
	close(this->pipeCmds[0]); // close reading in pipeCmds
	bzero(msgPipe, MSG_PIPE_LEN);
	nbytesread = read(this->pipeMeas[0], msgPipe, MSG_PIPE_LEN);
	if (nbytesread <= 0) {
		perror("No actual control signal written on House named pipe");
		return -2;
	}
	if (StringUtils_toULong(msgPipe, &(this->profilesFileCurrentLine)) < 0) {
		Debug_out(DEBUG, "Value '%s' not accepted\n", msgPipe);
		return -2;
	}
	Debug_out(DEBUG, "[COMMANDS] FROM PIPE [Meas] : %lu\n", this->profilesFileCurrentLine);

	/* control signal */
	if (control != this->profilesFileCurrentLine) {
		Debug_out(DEBUG, "Commanding with control '%ld' while actual control is '%ld'\n", control, this->profilesFileCurrentLine);
		return -2;
	}

	/* battery */
	this->batteryChargeCmd = commands[CMDS_BATTERY];

	/* phev */
	if (this->isSubstation || !this->PHEV) {
		this->phevChargeCmd = 0;
	}
	else {
		this->phevChargeCmd = commands[CMDS_PHEV];
	}

	Debug_out(DEBUG, "Command '%ld %lg %lg' registered\n", control, commands[CMDS_BATTERY], commands[CMDS_PHEV]);

	/* write commands to other server via named pipe */
	bzero(msgPipe, MSG_PIPE_LEN);
	sprintf(msgPipe, "%lg %lg\n", this->batteryChargeCmd, this->phevChargeCmd);
	nbytesread = write(this->pipeCmds[1], msgPipe, MSG_PIPE_LEN);
	if (nbytesread <= 0) {
		perror("Problems in writing battery and phev commands on House named pipe");
		return -2;
	}
	Debug_out(DEBUG, "[COMMANDS] TO PIPE [Cmds] : %lg %lg\n", this->batteryChargeCmd, this->phevChargeCmd);

	/* ack to command */
	return 0;
}






Server * Server_new(const char *name, const char * address, int port, int controlStopServer, int maxMsgSize, int maxAccConnAtTime)
{
	Debug_assert(DEBUG, name != NULL, "name == NULL\n");
	Server * new = (Server *) calloc(1, sizeof(Server));
	new->name = StringUtils_clone(name);
	if (address != NULL) {
		new->address = StringUtils_clone(address);
	}
	new->port = port;
	new->controlStopServer = controlStopServer;
	new->maxMsgSize = maxMsgSize;
	new->maxAcceptedConnAtTime = maxAccConnAtTime;
	return new;
}

void Server_free(Server **_this)
{
	Debug_assert(DEBUG, *_this != NULL, "*_this == NULL\n");
	free(*_this);
	*_this = NULL;
}

