#include <House.h>
#include <Debug.h>
#include <StringUtils.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <strings.h>



#define DEBUG "House"


static int _Server_openSocketAtPort(Server *_this);

static void * _Server_get_in_addr(struct sockaddr *sa);



pid_t House_startMeasurements(House *_this, FILE *stream)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	pid_t childProcess;

	childProcess = fork();

	/* Child code */
	if (childProcess == 0) {
		/* child is server always accepting requests */
		int fncres;
		struct sockaddr_storage their_addr; // connector's address information
		socklen_t sin_size;
		pid_t server_pid = getpid(); // pid of server
		int new_fd;
		char s[INET6_ADDRSTRLEN];
		double measurements[MEAS_NUMBER];

		/* Open server's socket */
		int sock = _Server_openSocketAtPort(_this->measurementsConsumer);
		if (sock == -1) {
			exit(1);
		}
		Debug_out(DEBUG, "[%s] Opened server at port %d\n", _this->measurementsConsumer->name, _this->measurementsConsumer->port);

		/* accept and close when control signal STOP is read */
		Debug_out(DEBUG, "[%s] Accepting request on %d\n", _this->measurementsConsumer->name, _this->measurementsConsumer->port);


		sin_size = sizeof(their_addr);

		new_fd = accept(sock, (struct sockaddr *)&their_addr, &sin_size);

		if (new_fd == -1) {
			Debug_out(DEBUG, "[%s] ERROR accepting\n", _this->measurementsConsumer->name);
			perror("");
			fflush(NULL);
			close(sock);
			exit(1);
		}

		inet_ntop(their_addr.ss_family,
				  _Server_get_in_addr((struct sockaddr *)&their_addr),
				  s, sizeof(s));
		Debug_out(DEBUG, "[%s] Connection from %s\n", _this->measurementsConsumer->name, s);

		while (1) {
			int numbytes, i;
			static long int controlsignal;
			static long int parsingRes;

			/* reads a message */
			numbytes = recv(new_fd, (char *)&controlsignal, sizeof(long int), 0);
			if (numbytes <= 0) { // client has closed connection
				Debug_out(DEBUG, "[%s] Client has closed its connection\n", _this->measurementsConsumer->name);
				break;
			}
			Debug_out(DEBUG, "[%s] Received message '%ld'\n", _this->measurementsConsumer->name, controlsignal);

			/* check if it is stop command */
			if (controlsignal == _this->commandsConsumer->controlStopServer) {
				close(new_fd);
				break;
			}

			/* take corresponding profile from file */
			parsingRes = House_measurementsReqsFromConsumer(_this, controlsignal, measurements);
			if (parsingRes == -2) {
				fprintf(stderr, "[%s] ERROR in reading measurements at control %ld\n", _this->measurementsConsumer->name, controlsignal);
				break;
			}
			else if (parsingRes == -1) {
				fprintf(stderr, "[%s] Server terminated profiles reading\n", _this->measurementsConsumer->name);
				break;
			}

			/* send answer back with measurements */
			numbytes = send(new_fd, (char *)&parsingRes, sizeof(long int), 0);
			if (numbytes <= 0) {
				fprintf(stderr, "[%s] ERROR in writing measurements to client\n", _this->measurementsConsumer->name);
				close(new_fd);
				close(sock);
				exit(1);
			}
			for (i = 0; i < MEAS_NUMBER; i++) {
				numbytes = send(new_fd, (char *)&measurements[i], sizeof(double), 0);
				if (numbytes <= 0) {
					fprintf(stderr, "[%s] ERROR in writing measurements to client\n", _this->measurementsConsumer->name);
					close(new_fd);
					close(sock);
					exit(1);
				}
			}
		}

		/* server shuts down */
		close(new_fd);
		close(sock);
		Debug_out(DEBUG, "[%s] Closed server at port %d\n", _this->measurementsConsumer->name, _this->measurementsConsumer->port);
		exit(0);
	}
	else if (childProcess < 0) { /* Error in fork() */
		Debug_out(DEBUG, "[%s] Error in creating server\n", _this->measurementsConsumer->name);
		perror("");
		fflush(NULL);
		return -1;
	}
	
	return childProcess;
}




pid_t House_startCommands(House *_this, FILE *stream)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	pid_t childProcess;

	childProcess = fork();

	/* Child code */
	if (childProcess == 0) {
		/* child is server always accepting requests */
		int fncres;
		struct sockaddr_storage their_addr; // connector's address information
		socklen_t sin_size;
		pid_t server_pid = getpid(); // pid of server
		int new_fd;
		char s[INET6_ADDRSTRLEN];
		static double commands[CMDS_NUMBER];

		/* Open server's socket */
		int sock = _Server_openSocketAtPort(_this->commandsConsumer);
		if (sock == -1) {
			exit(1);
		}
		Debug_out(DEBUG, "[%s] Opened server at port %d\n", _this->commandsConsumer->name, _this->commandsConsumer->port);

		/* accept and close when control signal STOP is read */
		Debug_out(DEBUG, "[%s] Accepting request on %d\n", _this->commandsConsumer->name, _this->commandsConsumer->port);


		sin_size = sizeof(their_addr);

		new_fd = accept(sock, (struct sockaddr *)&their_addr, &sin_size);

		if (new_fd == -1) {
			Debug_out(DEBUG, "[%s] ERROR accepting\n", _this->commandsConsumer->name);
			perror("");
			fflush(NULL);
			close(sock);
			exit(1);
		}

		inet_ntop(their_addr.ss_family,
				  _Server_get_in_addr((struct sockaddr *)&their_addr),
				  s, sizeof(s));
		Debug_out(DEBUG, "[%s] Connection from %s\n", _this->commandsConsumer->name, s);

		while (1) {
			int numbytes, i;
			static long int controlsignal;
			static long int parsingRes;

			/* reads a message with two commands */
			numbytes = recv(new_fd, (char *)&controlsignal, sizeof(long int), 0);
			if (numbytes <= 0) { // client has closed connection
				Debug_out(DEBUG, "[%s] Client has closed its connection\n", _this->measurementsConsumer->name);
				break;
			}
			Debug_out(DEBUG, "[%s] Received message '%ld'\n", _this->commandsConsumer->name, controlsignal);
			/* receiving command for battery */
			numbytes = recv(new_fd, (char *)&commands[CMDS_BATTERY], sizeof(double), 0);
			if (numbytes <= 0) { // client has closed connection
				Debug_out(DEBUG, "[%s] Client has closed its connection\n", _this->measurementsConsumer->name);
				break;
			} else if (controlsignal == _this->commandsConsumer->controlStopServer) {
				close(new_fd);
				break;
			}
			Debug_out(DEBUG, "[%s] Received message '%lg'\n", _this->commandsConsumer->name, commands[CMDS_BATTERY]);
			/* receiving command for PHEV */
			numbytes = recv(new_fd, (char *)&commands[CMDS_PHEV], sizeof(double), 0);
			if (numbytes <= 0) { // client has closed connection
				Debug_out(DEBUG, "[%s] Client has closed its connection\n", _this->measurementsConsumer->name);
				break;
			}
			Debug_out(DEBUG, "[%s] Received message '%lg'\n", _this->commandsConsumer->name, commands[CMDS_PHEV]);

			/* update current status */
			parsingRes = House_commandsFromConsumer(_this, controlsignal, commands);
			if (parsingRes == -2) {
				fprintf(stderr, "[%s] ERROR in reading measurements at control %ld\n", _this->commandsConsumer->name, controlsignal);
				close(new_fd);
				close(sock);
				exit(1);
			}

			/* send ack back */
			numbytes = send(new_fd, (char *)&parsingRes, sizeof(long int), 0);
			if (numbytes <= 0) {
				fprintf(stderr, "[%s] ERROR in sending ack to client\n", _this->commandsConsumer->name);
				close(new_fd);
				close(sock);
				exit(1);
			}
		}

		/* server shuts down */
		close(new_fd);
		close(sock);
		Debug_out(DEBUG, "[%s] Closed server at port %d\n", _this->commandsConsumer->name, _this->commandsConsumer->port);
		exit(0);
	}
	else if (childProcess < 0) { /* Error in fork() */
		Debug_out(DEBUG, "[%s] Error in creating server\n", _this->commandsConsumer->name);
		perror("");
		fflush(NULL);
		return -1;
	}
	
	return childProcess;
}







/* Returns socket fd if everything is OK, -1 otherwise */
static int _Server_openSocketAtPort(Server *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	struct sockaddr_in sin;

	/* Creates socket */
	int sock = socket(AF_INET, SOCK_STREAM, 0);
	if(sock == -1) {
		fprintf(stderr, "[%s] Error opening socket\n", _this->name);
		perror("");
		return -1;
	}
	sin.sin_port = htons(_this->port);
	sin.sin_addr.s_addr = 0;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_family = AF_INET;
	/* Binds socket */
	if(bind(sock, (struct sockaddr *)&sin,sizeof(struct sockaddr_in) ) == -1) {
		fprintf(stderr, "[%s] Error binding socket at port %d\n", _this->name, _this->port);
		perror("");
		return -1;
	}
	/* Listens on socket */
	if (listen(sock, _this->maxAcceptedConnAtTime) == -1) {
		fprintf(stderr, "[%s] Error listening on soket at port %d\n", _this->name, _this->port);
		perror("");
		return -1;
	}

	return sock;
}



// get sockaddr, IPv4 or IPv6:
void * _Server_get_in_addr(struct sockaddr *sa)
{
	if (sa->sa_family == AF_INET) {
		return &(((struct sockaddr_in*)sa)->sin_addr);
	}

	return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

