#ifndef __HASHTABLE__
#define __HASHTABLE__

#include <stddef.h>
#include <stdio.h>

/**
	A generic hash-table: both keys and values are generic pointers (void*), thus
	they can point to instances of arbitrary structures.
**/

typedef struct HashTable_Entry HashTable_Entry;

const void* const HashTable_Entry_key(const HashTable_Entry* const _this);
void* const HashTable_Entry_value(const HashTable_Entry* const this);


struct HashTable;
typedef struct HashTable HashTable;

// Returns sizeof(HashTable) (as the type is incomplete, other modules cannot execute sizeof())
size_t HashTable_sizeof();

/**
	Create a new empty hashtable, whose key elements are keySize bytes long.
 */
HashTable* HashTable_new(size_t keySize);
//HashTable* HashTable_new_FAKE(size_t keySize, char* (*keyToString)(const void* keyP, char* str));

/**
	Returns the value associated to the key value pointed by key in HashTable t_his, or NULL if 
	no such value exists.
	
	Arguments:
	 - keyP: a pointer to the key. The function will search for
		an entry in HashTable t_his whose key is an identical copy of value pointed by keyP as key 
		(i.e., not the value of the pointer itself).

	 - valueP: the address of a pointer to the value. The function will store
		the returned value into *valueP (NULL if no pair exists). 

	The function returns *valueP.
 */
void* HashTable_get(const HashTable* const _this, const void* const key, void** valueP);


/** Returns the number of bytes occupied by a single entry of this hash-table (memory for the key included, 
	memory for the value EXCLUDED).
	In other words, it returns the number of bytes that a call fo HashTable_clear() would free-up 
	(the memory of the value of the cleared entry 
	must be freed by the caller, so it is not included in the computation).
*/
size_t HashTable_Entry_sizeof(const HashTable* const _this);


/**
	Add pair (*key, value) to HashTable t_his, removing any other
	(*key, value') pair.
	
	Arguments:
	 - key: a pointer to the key. The HastTable will store
		a copy of the value pointed by keyP as key (i.e., not the pointer).
		t_his means that a call to free(key) will have no destroying effects
		on the HashTable content.

	 - value: a pointer to the value. The HashTable will store
		the pointer itself as value (i.e., it will not duplicate the value
		in another memory location).
		
	Returns a pointer to the old value if exists, otherwise NULL				
 */
void* HashTable_put(HashTable* const _this, const void* const key, void* const value);


/**
	Delete the pair having key *key from HashTable t_his, if one exists.
	
	The function looks for a key value which is an identical copy
	of the value pointed by key (i.e., *key, not the pointer itself).	

	Returns a pointer to the old value of 'key' if exists (otherwise NULL)	
 */
void* HashTable_clear(HashTable* const _this, const void* const key);


/** 	Return the number of entries in t_his HashTable */
unsigned long HashTable_size(const HashTable* const _this);

/**
	Free the memory allocates for the HashTable t_his and sets *t_his to NULL.
	In particular:
	 - the memory storing the key values is freed (as key values are copied)
	 - the memory storing pointers to values is freed
	 - the memory storing value is **not** freed (as the HashTable stores pointers
	   to values, not the values themselves).
 */
void HashTable_free(HashTable** const _this);

void HashTable_print(FILE* stream, const HashTable* const _this, char* (*keyToString)(const void* keyP, char* str), char* (*valueToString)(const void* valueP, char* str), unsigned int keyMaxStrLen, unsigned int valueMaxStrLen);

char* HashTable_dumpToString(char* string, const HashTable* const _this, char* (*keyToString)(const void* keyP, char* str), char* (*valueToString)(const void* valueP, char* str));

// HashTable Iterator
struct HashTableIterator;
typedef struct HashTableIterator HashTableIterator;

/** Create an Iterator for t_his HashTable */
HashTableIterator* HashTableIterator_new(const HashTable* const _this);

/** Return 1 if a call to HashTableIterator_next(t_his) will return an entry, 0 otherwise */
int HashTableIterator_hasNext(const HashTableIterator* const _this);

/** Return the next entry of the HashTable which 't_his' is iterating on, 
	NULL if no more entries exist */
HashTable_Entry* HashTableIterator_next(HashTableIterator* const _this);

void HashTableIterator_rewind(HashTableIterator* this);

/** Free memory used by t_his iterator. Also sets *t_his to NULL. */
void HashTableIterator_free(HashTableIterator** const _this);

// Returns sizeof(HashTableIterator) (as the type is incomplete, other modules cannot execute sizeof())
size_t HashTableIterator_sizeof();

#endif

