#include <stdio.h>
#include <stdlib.h>

#include <Debug.h>
#include <HashSet.h>
#include <HashTable.h>

/**
Hash-set based on HashTable
**/

struct HashSet {
	HashTable* hTable;
	char* fakeValuePtr;
};

/** Returns the number of bytes occupied by a single entry of this hash-set (memory for the key included.
	In other words, it returns the number of bytes that a call fo HashTable_clear() would free-up.
*/
size_t HashSet_Entry_sizeof(const HashSet* const this) {
	return HashTable_Entry_sizeof(this->hTable);
}


HashSet* HashSet_new(size_t keySize) {
	HashTable* hTable = HashTable_new(keySize);
	HashSet* new = (HashSet*)calloc(1, sizeof(HashSet));
	new->hTable = hTable;
	new->fakeValuePtr = (char*)calloc(1, sizeof(char));
	return new;
}

int HashSet_contains(const HashSet* const this, const void* const key) {
	if (!this) {
		fprintf(stderr, "HashSet_contains(): this cannot be null.\n");
		return 0;
	}
	void* valueP = NULL;
	valueP = HashTable_get(this->hTable, key, &valueP);
	return (valueP != NULL);
}

void HashSet_add(HashSet* const this, const void* const key) {
	HashTable_put(this->hTable, key, this->fakeValuePtr);
}


void HashSet_remove(HashSet* const this, const void* const key) {
	HashTable_clear(this->hTable, key);
}

unsigned long HashSet_size(const HashSet* const this) {
	return HashTable_size(this->hTable);
}

void HashSet_free(HashSet** const thisP) {	
	Debug_assert(DEBUG_ALWAYS, thisP != NULL, "thisP == NULL\n");
	HashSet* this = *thisP;
	HashTable_free(&this->hTable);
	free( this->fakeValuePtr );
	free(this);
	*thisP = NULL;
}






// HashSet Iterator
struct HashSetIterator {
	HashTableIterator* hTableIterator;
};


/** Create an Iterator for this HashSet */
HashSetIterator* HashSetIterator_new(const HashSet* const this) {
	HashSetIterator* it = (HashSetIterator*)calloc(1, sizeof(HashSetIterator));	
	it->hTableIterator = HashTableIterator_new(this->hTable);
	return it;
}	

/** Return 1 if a call to HashSetIterator_next(this) will return an entry, 0 otherwise */
int HashSetIterator_hasNext(const HashSetIterator* const this) {
	return HashTableIterator_hasNext(this->hTableIterator);
}

/** Return the next element of the HashSet which 'this' is iterating on, 
	NULL if no more elements exist */
const void* HashSetIterator_next(HashSetIterator* const this) {
	HashTable_Entry* e = HashTableIterator_next(this->hTableIterator);
	if (e != NULL) return (const void*)(HashTable_Entry_key(e));
	else return NULL;
}


/** Free memory used by this iterator. Also sets *this to NULL. */
void HashSetIterator_free(HashSetIterator** const thisP) {
	Debug_assert(DEBUG_ALWAYS, thisP != NULL, "thisP == NULL\n");
	HashSetIterator* this = *thisP;
	HashTableIterator_free(& this->hTableIterator );
	free(this);
	*thisP = NULL;
}

/*
int main() {
	HashSet* s = HashSet_new(sizeof(unsigned long));
	unsigned long* memory = calloc(10, sizeof(unsigned long));
	memory[0] = 10;
	memory[1] = 20;
	memory[2] = 30;
	int max = 2;
	for (int i=0; i<=max; i++) {
		HashSet_add(s, &memory[i]);
	}
	HashSetIterator* it = HashSetIterator_new(s);
	const void* e = NULL;
	printf("Set:\n");	
	while( (e = HashSetIterator_next(it)) != NULL) {		
		printf("%lu ", *((unsigned long*)e));
	}
	printf("\n");
}*/

