#include "HashTable.h"
#include <stdio.h>
#include <stdlib.h>
#include "uthash/src/uthash.h"

#include <Debug.h>

#define HASHTABLE_DEBUG "HashTable"

/**
	A generic hash-table: both keys and values are generic pointers (void*), thus
	they can point to instances of arbitrary structures.
**/

struct HashTable_Entry {
	void* key;
	void* value;
	UT_hash_handle hh; /* makes this structure hashable */
};

const void* const HashTable_Entry_key(const HashTable_Entry* const this) {
	assert(this != NULL);
	return this->key;
}
void* const HashTable_Entry_value(const HashTable_Entry* const this) {
	assert(this != NULL);
	return this->value;
}



void HashTable_Entry_free(HashTable_Entry** const thisP) {
	HashTable_Entry* this = *thisP;
	if (this == NULL) return;
	free(this->key);
	free(this);
	*thisP = NULL;
}

struct HashTable {
	size_t keySize;	
	HashTable_Entry* entries;
};

size_t HashTable_Entry_sizeof(const HashTable* const this) {
	return sizeof(HashTable_Entry) + this->keySize;
}
size_t HashTable_sizeof() {
	return sizeof(HashTable);
} 






/**
	Create a new empty hashtable, whose key elements are keySize bytes long.
 */
HashTable* HashTable_new(size_t keySize) {
	if (keySize <= 0) {
		fprintf(stderr, "HashTable_new(): keySize must be > 0\n");
		return NULL;
	}
	HashTable* new = (HashTable*)calloc(1, sizeof(HashTable));
	new->entries = NULL;
	new->keySize = keySize;
	return new;
}


/**
	Returns the value associated to the key value pointed by key in HashTable this, or NULL if 
	no such value exists.
	
	Arguments:
	 - keyP: a pointer to the key. The function will search for
		an entry in HashTable this whose key is an identical copy of value pointed by keyP as key 
		(i.e., not the value of the pointer itself).

	 - valueP: the address of a pointer to the value. The function will store
		the returned value into *valueP (NULL if no pair exists). 

	The function returns *valueP.
 */
void* HashTable_get(const HashTable* const this, const void* const keyP, void** valueP) {
	assert(this != NULL); 
	assert(keyP != NULL);	
	assert(valueP != NULL);

	HashTable_Entry* result = NULL;

	HASH_FIND(hh, this->entries, keyP, this->keySize, result);
	if (result == NULL) {
		*valueP = NULL;
	} else {
		*valueP = result->value;
	}
	return *valueP;
}


/**
	Add pair (*key, value) to HashTable this, removing any other
	(*key, value') pair.
	
	Arguments:
	 - keyP: a pointer to the key. The HastTable will store
		a copy of the value pointed by keyP as key (i.e., not the pointer).
		This means that a call to free(key) will have no destroying effects
		on the HashTable content.

	 - value: a pointer to the value. The HashTable will store
		the pointer itself as value (i.e., it will not duplicate the value
		in another memory location).		
		
	Returns a pointer to the old value if exists, otherwise NULL
 */
void* HashTable_put(HashTable* const this, const void* const keyP, void* const value) {
	assert(this != NULL);
	assert(keyP != NULL);
	assert(value != NULL);	
	HashTable_Entry* entry = NULL;
	HASH_FIND(hh, this->entries, keyP, this->keySize, entry);

	if (entry == NULL) {
		entry = (HashTable_Entry*)calloc(1, sizeof(HashTable_Entry));
		entry->key = calloc(1, this->keySize);
		memcpy(entry->key, keyP, this->keySize);
		entry->value = value;
		HASH_ADD_KEYPTR(hh, this->entries, entry->key, this->keySize, entry);
		//fprintf(stderr, "Adding %p --> %p\n", entry->key, entry->value);
		return NULL;
	}
	else {
		void* oldValue = entry->value;
		entry->value = value;
		//fprintf(stderr, "Changing %p --> %p\n", entry->key, entry->value);		
		return oldValue;
	}
}


/**
	Delete the pair having key *key from HashTable this, if one exists.
	
	The function looks for a key value which is an identical copy
	of the value pointed by key (i.e., *key, not the pointer itself).	
	
	Returns a pointer to the old value of 'key' if exists (otherwise NULL)
 */
void* HashTable_clear(HashTable* const this, const void* const key) {
	assert(this != NULL);
	assert(key != NULL);
	HashTable_Entry* result = NULL;
	void* oldValue = NULL;
	HASH_FIND(hh, this->entries, key, this->keySize, result);
	if (result != NULL) {
		oldValue = result->value;
		HASH_DEL(this->entries, result);
		HashTable_Entry_free(&result);		
	}
	return oldValue;
}


/** 	Return the number of entries in this HashTable */
unsigned long HashTable_size(const HashTable* const this) {
	return HASH_COUNT(this->entries);
}

/**
	Free the memory allocates for the HashTable this and sets *this to NULL.
	In particular:
	 - the memory storing the key values is freed (as key values are copied)
	 - the memory storing pointers to values is freed
	 - the memory storing value is **not** freed (as the HashTable stores pointers
	   to values, not the values themselves).
 */
void HashTable_free(HashTable** const thisP) {
	Debug_assert(HASHTABLE_DEBUG, thisP != NULL, "thisP == NULL");
	HashTable* this = *thisP;
	if (this == NULL) return;

	//if (HashTable_size(*this) > 0) HASH_CLEAR(hh, (*this)->entries);	
	HashTableIterator* it = HashTableIterator_new(this);
	HashTable_Entry* entry = NULL;
	while ((entry = HashTableIterator_next(it)) != NULL) {
		HASH_DEL(this->entries, entry);
		HashTable_Entry_free(&entry);
	}
	HashTableIterator_free(&it);
	Debug_assert(HASHTABLE_DEBUG, HashTable_size(this) == 0, 
		"hTable still has %lu entries\n", HashTable_size(this));
	
	free(this);
	*thisP = NULL;
}



void HashTable_print(FILE* stream, const HashTable* const this, char* (*keyToString)(const void* keyP, char* str), char* (*valueToString)(const void* valueP, char* str), unsigned int keyMaxStrLen, unsigned int valueMaxStrLen) {
	assert(stream != NULL);
	assert(this != NULL);
	assert(keyToString != NULL);
	assert(valueToString != NULL);
	
	fprintf(stream, "BEGIN (%lu elements)\n", HashTable_size(this));
	HashTableIterator* it = HashTableIterator_new(this);
	HashTable_Entry* e = NULL;
	char* strK = calloc(keyMaxStrLen+1, sizeof(char));
	char* strV = calloc(valueMaxStrLen+1, sizeof(char));	
	while ( (e = HashTableIterator_next(it)) != NULL) {
		fprintf(stream, "  '%s' (%p) --> '%s' (%p)\n", (*keyToString)(e->key, strK), e->key, (*valueToString)(e->value, strV), e->value);		
	}
	fprintf(stream, "END\n");
	free(strK);
	free(strV);
	HashTableIterator_free(&it);	
}

char* HashTable_dumpToString(char* string, const HashTable* const this, char* (*keyToString)(const void* keyP, char* str), char* (*valueToString)(const void* valueP, char* str)) {
	assert(string != NULL);
	assert(this != NULL);
	assert(keyToString != NULL);
	assert(valueToString != NULL);
	
	sprintf(&string[0], "BEGIN (%lu elements)\n", HashTable_size(this));
	HashTableIterator* it = HashTableIterator_new(this);
	HashTable_Entry* e = NULL;
	char* strK = calloc(10000, sizeof(char));
	char* strV = calloc(10000, sizeof(char));	
	while ( (e = HashTableIterator_next(it)) != NULL) {
		sprintf(&string[strlen(string)], "  '%s' (%p) --> '%s' (%p)\n", (*keyToString)(e->key, strK), e->key, (*valueToString)(e->value, strV), e->value);		
	}
	sprintf(&string[strlen(string)], "END\n");
	free(strK);
	free(strV);
	HashTableIterator_free(&it);	
	return string;
}




// HashTable Iterator
struct HashTableIterator {
	const HashTable* hTable;
	HashTable_Entry* nextEntry;
};
size_t HashTableIterator_sizeof() {
	return sizeof(HashTableIterator);
}


/** Create an Iterator for this HashTable */
HashTableIterator* HashTableIterator_new(const HashTable* const this) {
	assert(this != NULL);
	HashTableIterator* it = (HashTableIterator*)calloc(1, sizeof(HashTableIterator));	
	it->hTable = this;
	it->nextEntry = this->entries;
	return it;
}	

void HashTableIterator_rewind(HashTableIterator* const this) {
	this->nextEntry = this->hTable->entries;
}

/** Return 1 if a call to HashTableIterator_next(this) will return an entry, 0 otherwise */
int HashTableIterator_hasNext(const HashTableIterator* const this) {
	assert(this != NULL);	
	return this->nextEntry != NULL;
}

/** Return the next entry of the HashTable which 'this' is iterating on, 
	NULL if no more entries exist */
HashTable_Entry* HashTableIterator_next(HashTableIterator* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");	
	HashTable_Entry* result = this->nextEntry;
	if (this->nextEntry != NULL) {
		this->nextEntry = (this->nextEntry->hh).next;
	}
	return result;
}

/** Free memory used by this iterator. Also sets *this to NULL. */
void HashTableIterator_free(HashTableIterator** const thisP) {
	Debug_assert(DEBUG_ALWAYS, thisP != NULL, "thisP == NULL\n");
	HashTableIterator* this = *thisP;
	free(this);
	*thisP = NULL;
}




