#include <stdio.h>
#include <stdlib.h>

#include <Debug.h>
#include <Numbers.h>

#include <Range_Double.h>

#define DEBUG "Range_Double"



struct Range_Double {
	double min;
	double max;
	double step;
	
	// cache
	int size;
};

typedef enum {NORMAL, FLIPPING} It_Type;
struct Range_Double_Iterator {
	Range_Double* range;
	It_Type type;

	int hasNext;	
	double next;
	
	// Additional info for FLIPPING iterators
	double med;
	int absDelta;
	int goRight;	
	int minReached;
	int maxReached;
	
};


Range_Double* Range_Double_new(double min, double max, double step) {
	Debug_assert(DEBUG_ALWAYS, max >= min, "max = %lf < %lf = min\n", max, min);
	Debug_assert(DEBUG_ALWAYS, step > 0, "step = %lf <= '\n", step);
	Range_Double* new = calloc(1, sizeof(Range_Double));
	Debug_assert(DEBUG_ALWAYS, new != NULL, "new == NULL\n");
	new->min = min;
	new->max = max;
	new->step = step;
	new->size = -1;
	return new;
}
double Range_Double_min(const Range_Double* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");
	return this->min;
}
double Range_Double_max(const Range_Double* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");	
	return this->max;
}
double Range_Double_step(const Range_Double* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");	
	return this->step;
}
void Range_Double_free(Range_Double** thisP) {
	Debug_assert(DEBUG_ALWAYS, thisP != NULL, "thisP == NULL\n");
	Range_Double* this = *thisP;
	if (this == NULL) return;
	free(this);
	*thisP = NULL;
}

int Range_Double_size(Range_Double* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");
	if (this->size < 0) {
		this->size = 0;
		for(double i = this->min; Numbers_approxLE(i, this->max); i += this->step) {
			this->size++;
		}
	}
	return this->size;
}







Range_Double_Iterator* __Range_Double_Iterator_new(Range_Double* const range, It_Type type) {
	Debug_assert(DEBUG_ALWAYS, range != NULL, "range == NULL");
	Range_Double_Iterator* new = calloc(1, sizeof(Range_Double_Iterator));
	Debug_assert(DEBUG_ALWAYS, new != NULL, "new == NULL");
	new->range = range;
	new->type = type;
	new->absDelta = 0;
	new->med = 0;
	new->goRight = 0;
	new->minReached = 0;
	new->maxReached = 0;	
	Range_Double_Iterator_rewind(new);
	return new;
}
Range_Double_Iterator* Range_Double_Iterator_new(Range_Double* const range) {
	return __Range_Double_Iterator_new(range, NORMAL);
}
Range_Double_Iterator* Range_Double_Iterator_new_flipping(
	Range_Double* const range) {
		return __Range_Double_Iterator_new(range, FLIPPING);
}




int Range_Double_Iterator_hasNext(Range_Double_Iterator* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");	
	return this->hasNext;
}	

void next_flipping(Range_Double_Iterator* const this) {
	//Debug_out(DEBUG, "Start next_flipping()\n");
	if ( !this->goRight) {
		// I just went left
		if (Numbers_approxLE(this->next, Range_Double_min(this->range))) {
			Debug_out(DEBUG, "MIN REACHED\n");
			this->minReached = 1;	// remember if min was reached	
		}
		
		// prepare to go right
		this->absDelta++;		
		if (!this->maxReached) {
			this->next = this->med + (this->absDelta * Range_Double_step(this->range));
			/*Debug_assert(DEBUG, this->next <= Range_Double_max(this->range), 
				"this->next > this->range->max\n");*/
			this->goRight = 1;
		} else {
			// keep going left
			this->next = this->med - (this->absDelta * Range_Double_step(this->range));
			/*Debug_assert(DEBUG, this->next >= Range_Double_min(this->range), 
				"this->next < this->range->min\n");*/
		}
	} else {
		// I just went right
		if (Numbers_approxGE(this->next, Range_Double_max(this->range))) {
			Debug_out(DEBUG, "MAX REACHED\n");
			this->maxReached = 1;	// remember if max was reached
		}
		
		// prepare to go left
		if (!this->minReached) {
			this->next = this->med - (this->absDelta * Range_Double_step(this->range));
			/*Debug_assert(DEBUG, this->next >= Range_Double_min(this->range), 
				"this->next < this->range->min\n");*/
			this->goRight = 0;
		} else {
			// keep going right
			this->absDelta++;
			this->next = this->med + (this->absDelta * Range_Double_step(this->range));
			/*Debug_assert(DEBUG, this->next <= Range_Double_max(this->range), 
				"this->next > this->range->max\n");*/
		}
	}
	this->hasNext = (!this->minReached || !this->maxReached);
	//Debug_out(DEBUG, "End next_flipping()\n");	
}


int Range_Double_Iterator_next(Range_Double_Iterator* const this, double* result) {
	if (!Range_Double_Iterator_hasNext(this)) return 0;
	Debug_assert(DEBUG_ALWAYS, result != NULL, "result == NULL\n");
	*result = this->next;
	
	switch(this->type) {
		case NORMAL :
			this->next += Range_Double_step(this->range);
			this->hasNext = 
				Numbers_approxLE(this->next, Range_Double_max(this->range));
			break;
			
		case FLIPPING :
			next_flipping(this);
			break;
	}
	return 1;
}
void Range_Double_Iterator_rewind(Range_Double_Iterator* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");	

	//Debug_out(DEBUG, "Start rewind()\n");
	this->hasNext = 1;
	switch(this->type) {
		case NORMAL :
			this->next = Range_Double_min(this->range);
			break;
			
		case FLIPPING :
			this->absDelta = 0;
			this->med = Range_Double_min(this->range) + 
				Range_Double_step(this->range) * (Range_Double_size(this->range)/2);
			//Debug_out(DEBUG, " - this->med = %lf\n", this->med);
			this->next = this->med;
			this->goRight = 0;
			this->minReached = 0;
			this->maxReached = 0;
			break;
		default:
			Debug_assert(DEBUG, 0, "Iterator type %d unknown\n", this->type);
	}
	//Debug_out(DEBUG, "End rewind()\n");
}
void Range_Double_Iterator_free(Range_Double_Iterator** thisP) {
	Debug_assert(DEBUG_ALWAYS, thisP != NULL, "thisP == NULL\n");
	Range_Double_Iterator* this = *thisP;
	if (this == NULL) return;
	free(this);
	*thisP = NULL;
}



