#! /usr/bin/env python

#
#  Convexer 2.0
#
#      'Constraint' class to store the result of implications' convertions
#
#      NEW: Supports multiple input line with backslash
#      NEW: Supports formulas among equations. Forbidden: the not equal operator.
#      NEW: Supports exponential format real numbers
#
#      authors: 
#           Federico Mari <federicomari@gmail.com> and 
#           Igor Melatti <igor.melatti@gmail.com>
#

import sys

class Constraint:
    """
    A constraint has the following fields:
    - terms  : dictionary of coefficients, terms[variable_id] = coefficient
    - bounds : dictionary of bounds, bounds[variable_id] = [lower, upper]
    - rel    : relation in {EQ, LEQ, GEQ} resp for {=, <=, >=}
    - b      : right hand side of the constraint
    """
    def __init__(self):
        self.terms = {}
        self.bounds = {}
        self.rel = ''
        self.b = 0

    def __str__(self):
        """ string representation """
        s, first = "", True
        for term in self.terms:
            if self.terms[term] == 0: continue
            if first:
                if self.terms[term] == -1: s = s + "-"
                elif self.terms[term] != 1: s = s + str(self.terms[term]) + " "
                first = False
            else:
                if self.terms[term] == -1: s = s + "- "
                elif self.terms[term] < 0: s = s + "- " + str(-self.terms[term]) + " "
                elif self.terms[term] == 1: s = s + "+ "
                else: s = s + "+ " + str(self.terms[term]) + " "
            s = s + term + ' '
        if self.rel != '':
            s = s + ("=" if self.rel == 'EQ' else ("<=" if self.rel == 'LEQ' else ">=")) + ' '
        else:
            s = s + ', b: '
        s = s + str(self.b)
        return s

    def str_lhs(self):
        """ returns a string containint the Left Hand Side (LHS) of self """
        s = ""
        for term in self.terms:
            s = s + '+ ' + (str(self.terms[term]) if self.terms[term] > 0 else '(' + str(self.terms[term]) + ')') + ' ' + term
        return s

    def setTerm(self, var, coef):
        self.terms[var] = coef

    def addTerm(self, var, coef):
        if not var in self.terms:            
            self.setTerm(var, coef)
        else:
            self.terms[var] += coef

    def addTerms(self, dict, changeSign):
        """ add terms of dict to self, possibly changing sign of terms in dict """
        for term in dict:
            if term in self.terms:
                self.addTerm(term, -dict[term] if changeSign else dict[term])
            else:
                self.setTerm(term, -dict[term] if changeSign else dict[term])

    def getb(self):
        return self.b

    def setb(self, b):
        self.b = b

    def getBounds(self):
        return self.bounds

    def setBounds(self, bounds):
        self.bounds = bounds

    def update(self, other, changeSign):
        """ 
        Add the lhs (and rhs) of other to the lhs (and rhs) of self.
        Subtract if changeSign is True.
        """
        if self.rel != other.rel: return None
        self.addTerms(other.terms, changeSign)
        self.b += -other.b if changeSign else other.b
        return self

    def opposite(self):
        """ returns a new Constraint containing the negation of self """
        res = Constraint()
        if self.rel != '':
            res.b = -self.b
            if self.rel == 'LEQ':
                res.rel = 'GEQ'
            elif self.rel == 'GEQ':
                res.rel = 'LEQ'
            else: res.rel = 'EQ'
        for term in self.terms:
            res.terms[term] = -self.terms[term]
        return res

    def max(self):
        """ returns the maximum value attainable by the lhs """
        if len(self.terms) == 0: return 0
        res = 0
        for term in self.terms:
            res += self.terms[term] * (self.bounds[term][0] if self.terms[term] < 0 else self.bounds[term][1])
        return res

    # returns an Abstract Syntax Tree (AST) of the form [self.rel, lhs, rhs]
    def getAST(self):
        # relation
        ret = [self.rel]
        # lhs
        k = self.terms.keys()
        if len(k) == 0:
            ret.append([])
        elif len(k) == 1:
            ret.append(['MONOMIAL', {k[0]: self.terms[k[0]]}])
        elif len(k) == 2:
            ret.append(['PLUS', ['MONOMIAL', {k[0]: self.terms[k[0]]}], ['MONOMIAL', {k[1]: self.terms[k[1]]}]])
        else:
            lhs = ['PLUS', ['MONOMIAL', {k[0]: self.terms[k[0]]}], ['MONOMIAL', {k[1]: self.terms[k[1]]}]]
            for key in range(2,len(k)):
                lhs = ['PLUS', ['MONOMIAL', {k[key]: self.terms[k[key]]}], lhs]
            ret.append(lhs)
        # rhs
        ret.append(['CONST', self.b])
        return ret

    def convert_to_LEQ(self):
        res = []
        if self.rel == 'GEQ':
            res.append(self.opposite())
        elif self.rel == 'EQ':
            self.rel = 'LEQ'
            res.append(self)
            res.append(self.opposite())
            res[1].rel = 'LEQ'
        elif self.rel == 'LEQ':
            res.append(self)
        return res

#########


if (__name__=="__main__"):
    c = Constraint()
    print c.__dict__
    c.addTerms({'x1': 1, 'x2': 2, 'x3': 3}, False)
    c.rel = 'EQ'
    print c.terms
    print c.getAST()
    for con in c.convert_to_LEQ():
        print con, ", ", 
