#!/usr/bin/env python

"""
Convexer 2.0

guardedconstraint.py

    'GuardedConstraint' class to store implications inequalities,
     namely guarded constraints.

    Created by Federico Mari and Igor Melatti.
    For any bugs write to <mari@di.uniroma1.it> or <melatti@di.uniroma1.it>.
    Copyright (c) 2010 . All rights reserved.
"""

import sys
import constraint

class GuardedConstraint():
    """
    A guarded constraint is an implication of the form:
        y -> c:Constraint
    Thus, a GuardedConstraint has the following fields:
    - y : which is the lhs of the implication
    - c : which is the rhs of the implication
    """
    def __init__(self):
        self.pole = True
        self.y = ''
        self.c = constraint.Constraint()

    def __str__(self):
        """string representation"""
        s = ''
        if self.y != '':
            if self.pole:
                s = s + self.y + ' -> '
            else:
                s = s + "!" + self.y + ' -> '
        s = s + str(self.c)
        return s
    
    def setPole(self, pole):
        self.pole = pole

    def setLhs(self, y):
        self.y = y

    def setRhs(self, constr):
        self.c = constr
    
    def getAST(self):
        if self.pole:
            ret = ['IMPLIES']
        else:
            ret = ['NOT IMPLIES']
        if self.y != '':
            ret.append(self.y)
        ret.append(self.c.getAST())
        return ret
    
    def split(self):
        if isinstance(self.c, constraint.Constraint):
            if self.pole:
                return [[self.y], self.c]
            else:
                return [["!" + self.y], self.c]
        else:
            right = self.c.split()
            if self.pole:
                right[0].append(self.y)
            else:
                right[0].append("!" + self.y)
            return [right[0], right[1]]

    def convert_to_LEQ(self):
        res = []
        constrs = self.c.convert_to_LEQ() # first convert the rhs of c
        for con in constrs:
            gc = GuardedConstraint()
            gc.setPole(self.pole)
            gc.setLhs(self.y)
            gc.setRhs(con)
            res.append(gc)
        return res

#########



if __name__ == '__main__':
    c = constraint.Constraint()
    c.addTerm('x', 2)
    c.rel = "EQ"
    c.setb(1)
    gc = GuardedConstraint()
    gc.setRhs(c)
    gc.setLhs("y1")
    gc2 = GuardedConstraint()
    gc2.setRhs(gc)
    gc2.setLhs("y2")

    print c.getAST()
    print gc
    print gc2
    print gc2.getAST()
    print gc2.split()
    for con in gc2.convert_to_LEQ():
        print con, ", ", 
