#! /usr/bin/env python

from __future__ import with_statement

import sys
import os
import pdb

# Lex and Yacc
import lex
import yacc

from ..constr import constraint
from ..constr import guardedconstraint
from .. import printer

class Convex2:

    def __init__(self, options, printer):
        self.options = options
        self.printer = printer
        self.auxiliary = []

    def convex(self, ast):
        """
        /================================================\
        |       OLD                         NEW            |
        |--------------------------------------------------|
        |    y -> ax <= b     ==>    ax + (M - b)y <= M    | (1)
        |--------------------------------------------------|
        |    !y -> ax <= b    ==>    ax - (M - b)y <= b    | (2)
        \================================================/
        
        Supports FORMULAS and not convexing (only transform to guarded predicate)
        
        Function 'convex' traverses the Abstract Syntax Tree (ast) generated by the parser
        and convexes the represented implication following (1) or (2) if necessary
        """
        # An AST can begin with EQ, LEQ, GEQ, IMPIES, NOT IMPLIES, AND, OR, NOT EQ or NOT FORMULA
        res = []
        if ast[0] == 'EQ' or ast[0] == 'LEQ' or ast[0] == 'GEQ': # ast = [:constraint:]
            # Obtain a list of Constraints: two if ast[0] == 'EQ'; one ow.
            temp = self.printer.get_constraint(ast)
            res = self.convert_to_LEQ(temp)
        elif ast[0] == 'AND' or ast[0] == 'OR' or ast[0] == 'NOT EQ' or ast[0] == 'NOT FORMULA': # we have a formula, then we have to transform it
            if self.options.verbose:
                self.printer.output("formula:")
                self.printer.output_formula(ast)
                self.printer.output("")
            # 1. recursively eliminate negation
            f = self.negate(ast, True)
            if self.options.verbose:
                self.printer.output("- negated:")
                self.printer.output_formula(f)
                self.printer.output("")
            # 2. recursively eliminate OR
            g = self.massage_or(f, False)
            if self.options.verbose:
                self.printer.output("- without OR: ")
                self.printer.output_formula(g)            
                self.printer.output("")
            # Post: g is a combination of implications and ands
            # 3. convex the formula by transforming implications
            G = self.convex_single_formula(g)
            if self.options.verbose:
                self.printer.output("- final conj: ")
                for eq in G:
                    self.printer.output(eq)
            res = G
        elif ast[0] == 'IMPLIES': # ast = ['IMPLIES', 'y', [:constraint:]]
            if ast[2][0] == 'IMPLIES' or ast[2][0] == 'NOT IMPLIES' or ast[2][0] == 'AND' or ast[2][0] == 'OR' or ast[2][0] == 'NOT EQ' or ast[2][0] == 'NOT FORMULA': # we have an implication or a formula in the rhs -> transform it
                constraints = self.convex(ast[2]) # recursive call -> to obtain a conj of contraints in the rhs
            else: # we have a simple equation, then get a constraint and transform it in form LEQ
                con = self.printer.get_constraint(ast[2])
                constraints = self.convert_to_LEQ(con)
            for c in constraints:
                if self.options.convex:
                    c.setBounds(yacc.getBounds()) # bounds are stored in the yacc module during parsing
                    M = c.max()
                    b = c.getb()
                    if self.options.verbose:
                        self.printer.output(str(c) + ": M = " + str(M))
                        self.printer.output("--> " + c.str_lhs() + " + (M - b) " + ast[1] + " <= M")
                    c.setb(M)
                    c.setTerm(ast[1], M - b)
                    res.append(c)
                else:
                    gc = guardedconstraint.GuardedConstraint()
                    gc.setLhs(ast[1])      # the lhs of the implication, the 'y' above
                    gc.setRhs(c)
                    res.append(gc)
        elif ast[0] == 'NOT IMPLIES': # ast = ['NOT IMPLIES', 'y', [:constraint:]]
            if ast[2][0] == 'IMPLIES' or ast[2][0] == 'NOT IMPLIES' or ast[2][0] == 'AND' or ast[2][0] == 'OR' or ast[2][0] == 'NOT EQ' or ast[2][0] == 'NOT FORMULA': # we have a formula in the rhs -> transform it
                constraints = self.convex(ast[2]) # recursive call -> to obtain a conj of contraints in the rhs
            else: # we have a simple equation, then get a constraint and transform it in form LEQ
                con = self.printer.get_constraint(ast[2])
                constraints = self.convert_to_LEQ(con)
            for c in constraints:
                if self.options.convex:
                    c.setBounds(yacc.getBounds()) # bounds are stored in the yacc module during parsing
                    M = c.max()
                    b = c.getb()
                    if self.options.verbose:
                        self.printer.output(str(c) + ": M = " + str(M))
                        self.printer.output("--> " + c.str_lhs() + " - (M - b) " + ast[1] + " <= b")
                    c.setTerm(ast[1], -(M - b))
                    res.append(c)
                else:
                    gc = guardedconstraint.GuardedConstraint()
                    gc.setPole(False)
                    gc.setLhs(ast[1])      # the lhs of the implication, the 'y' negated above
                    gc.setRhs(c)
                    res.append(gc)
        else: print "Semantic Error (convex): " + str(ast)
        return res
    
    def convert_to_LEQ(self, c):
        res = []
        if isinstance(c, guardedconstraint.GuardedConstraint):
            constrs = self.convert_to_LEQ(c.c) # first convert the rhs of c
            for con in constrs:
                gc = guardedconstraint.GuardedConstraint()
                gc.setPole(c.pole)
                gc.setLhs(c.y)
                gc.setRhs(con)
                res.append(gc)
        elif isinstance(c, constraint.Constraint):
            if c.rel == 'GEQ':
                if self.options.verbose:
                    self.printer.output("GEQ -> LEQ")
                res.append(c.opposite())
            elif c.rel == 'EQ':
                if self.options.verbose:
                    self.printer.output("EQ -> 2 LEQ")
                c.rel = 'LEQ'
                res.append(c)
                res.append(c.opposite())
                res[1].rel = 'LEQ'
            elif c.rel == 'LEQ':
                res.append(c)
        else: 
            print "Semantic Error (convert_to_LEQ): " + str(c)
            sys.exit()
        return res

    def negate(self, f, polarity):
        """
        Walking on formulas (negating). It is an error to negate a constraint with equality. We do not support that.
        """
        # base case: constraint
        if f[0] == 'EQ' or f[0] == 'LEQ' or f[0] == 'GEQ':
            g = f
            if not polarity:
                if g[0] == 'EQ':
                    print "Semantic Error: Unable to do 'not equal' : " + str(f)
                    sys.exit()
                elif g[0] == 'LEQ': 
                    g[0] = 'GEQ'
                else:
                    g[0] = 'LEQ'
            return g
        # recursions
        if f[0] == 'NOT EQ':
            g = f[1]
            if polarity:
                if g[0] == 'EQ':
                    print "Semantic Error: Unable to do 'not equal' : " + str(f)
                    sys.exit
                elif g[0] == 'LEQ': 
                    g[0] = 'GEQ'
                else:
                    g[0] = 'LEQ'
            return g
        if f[0] == 'NOT FORMULA': # has one child only
            return self.negate(f[1], not polarity)
        if f[0] == 'AND' or f[0] == 'OR':
            label = f[0]
            if f[0] == 'AND' and not polarity: label = 'OR'
            if f[0] == 'OR' and not polarity: label = 'AND'
            return [label, self.negate(f[1], polarity), self.negate(f[2], polarity)]
        # useless!
        return [f[0], self.negate(f[1], polarity), self.negate(f[2], polarity)]
    
    def massage_or(self, f, IamInOr):
        """
        2. massage or
        Pre: - f hasn't 'NOT *' nodes, since they where eliminated by negation()
             - formula ::= formula AND formula
                       ::= formula OR formula
                       ::= equation
        Idea: we want to eliminate OR in the following way
                       a \lor b   \equiv  y_1 -> a  \land  y_2 -> b  \land  y_1 + y_2 \geq 1
        Recursive implementation:
             The flag IamInOr is up if we are walking into a chain of ORs
             equation: if IamInOr then we must add the implication and bring up the new var
                       else return the simple equation, as it is
             and:      recur in the two operands.
                       if IamInOr then we have something of type [AND, l, r]. Since we are in
                       OR, we must add the implication new_var -> [AND, l, r].
                       This means adding new_var -> l AND new_var -> r. The distribution is a
                       bit trickier, thus we delegate it to another recursive function
                       (distribute_implication_in_and)
             or:       recur in the two operands.
                       if not IamInOr then add the constraint of new variables
        """
        if f[0] == 'EQ' or f[0] == 'LEQ' or f[0] == 'GEQ': # f = [:constraint:]
            if IamInOr:
                # new name and update the global couter, and return the implication
                name = self.options.prefix_of_new_vars + str(self.options.varcounter)
                self.auxiliary.append(name)
                self.options.varcounter += 1
                # add bounds of new variable in Yacc module
                yacc.addVarBound(name, 0, 1) # name is a boolean variable
                return ['IMPLIES', name, f, [name]] # must bring up the new variable, as 4th field
            else:
                return f
        if f[0] == 'AND': # f = ['AND', formula, formula]
            recurT = self.massage_or(f[1], False)
            recurE = self.massage_or(f[2], False)
            if IamInOr:
                # new name and update the global couter, and return the implication
                name = self.options.prefix_of_new_vars + str(self.options.varcounter)
                self.options.varcounter += 1
                self.auxiliary.append(name)
                # add bounds of new variable in Yacc module
                yacc.addVarBound(name, 0, 1) # name is a boolean variable
                # distribute the implication among members of AND
                # print "ret: " + str(ret)
                recurT = self.distribute_implication_in_and(name, recurT)
                recurE = self.distribute_implication_in_and(name, recurE)
                # the last field with the newly used variables
                return ['AND', recurT, recurE, [name]]
            else:
                return ['AND', recurT, recurE]
        if f[0] == 'OR': # f = ['OR', formula, formula]
            f1 = self.massage_or(f[1], True)
            g1 = f1[0:len(f1)-1]
            f2 = self.massage_or(f[2], True)
            g2 = f2[0:len(f2)-1]
            if not IamInOr: # this is the first or, the outer one -> add the new constraint
                # fi[3] contains the list of new added variables, i in {1, 2}
                new_eq = self.ast_constraint(f1[3], f2[3])
                return ['AND', new_eq, ['AND', g1, g2]]
            return ['AND', g1, g2, f1[3] + f2[3]]
        # never be here!
        return f
    
    def distribute_implication_in_and(self, var, ast):
        """
        3. distribute_implication_in_and for OR elimination
        Given y and a conjunction, distributes y -> among all conj members
        y -> [AND, a, b]  \equiv  [AND, y -> a, y -> b]
        Returns [AND, y -> a, y -> b]
        Note: we can obtain y1 -> (y2 -> eq). In such a case we must convexify later.
        """
        # print "ast: " + str(ast)
        if ast[0] == 'AND':
            return ['AND', self.distribute_implication_in_and(var, ast[1]), self.distribute_implication_in_and(var, ast[2])]
        # else we have an equation or implication
        return ['IMPLIES', var, ast]
    
    def ast_constraint(self, l1, l2):
        """
        4. ast_constraint for OR elimination
        Given two disjoint lists [y_1, ..., y_k1] and [y_1, ..., y_k2] of 
        newly added variables (added for implication)
        returns the constraint y_1 + ... + y_k1 + y_1 + ... + y_k2 \geq 1
        """
        poly1 = self.ast_constraint_rec(l1)
        poly2 = self.ast_constraint_rec(l2)
        polynomial = ['PLUS', poly1, poly2]
        return ['GEQ', polynomial, ['CONST', 1]]
    
    def ast_constraint_rec(self, l):
        # case base: len(l) == 0 or len(l) == 1
        if len(l) == 1:
            return ['MONOMIAL', {l[0]: 1}]
        if len(l) == 2:
            return ['PLUS', ['MONOMIAL', {l[0]: 1}], ['MONOMIAL', {l[1]: 1}]]
        # induction: sum of l[0] and the recursion on the rest of the list
        return ['PLUS', ['MONOMIAL', {l[0]: 1}], self.ast_constraint_rec(l[1:])]
    
    def convex_single_formula(self, g):
        """
        Convexes a single formula with the following grammar:
            formula ::= formula AND formula
            formula ::= arrow
            arrow   ::= ID -> arrow
            arrow   ::= equation
        this is called after a massage to formulas...
        """
        eqs = self.convex_single_formula_rec(g) # this returns a list of ASTs each representing an equation
        ret = []
        for eq in eqs:
            con = self.printer.get_constraint(eq)
            ret = ret + self.convert_to_LEQ(con)
        return ret
    
    def convex_single_formula_rec(self, g):
        """
        Obtains a list of ASTs each representing an equation in g
        """
        if g[0] == 'AND':
            return self.convex_single_formula_rec(g[1]) + self.convex_single_formula_rec(g[2])
        if g[0] == 'IMPLIES':
            # g = ['IMPLIES', 'var', x].... 
            # where x = g[2] could be:
            #   1. ['IMPLIES', 'anothervar', y] or 
            #   2. an equation
            if g[2][0] == 'IMPLIES': # we must recur, and the result is a list of equations
                equation_box = self.convex_single_formula_rec(g[2])
            else: # we have only an equation to be implied
                equation_box = [g[2]]
            if self.options.verbose:
                self.printer.output("  euqation_box:  " + str(equation_box))
            ret = []
            for eq in equation_box:
                if self.options.verbose:
                    self.printer.output("  *  " + str(g[1]) + " -> " + str(eq))
                con = self.convex_return_ast(['IMPLIES', g[1], eq])
                if self.options.verbose:
                    for c in con:
                        self.printer.output("  +  " + str(c))
                ret = ret + con
            return ret
        return [g]
    
    def convex_return_ast(self, ast):
        """
        We need convex to return an AST in convex_single_formula
        """
        res = []
        if ast[0] == 'IMPLIES': # ast = ['IMPLIES', 'y', [:constraint:]]
            constr = ast[2]
            partial = []
            con = self.printer.get_constraint(constr)
            partial = self.convert_to_LEQ(con)
            # for each constraint generated, apply (1) above
            for c in partial:
                if self.options.convex:
                    c.setBounds(yacc.getBounds()) # bounds are stored in the yacc module during parsing
                    M = c.max()
                    b = c.getb()
                    if self.options.verbose:
                        self.printer.output(str(c) + ": M = " + str(M))
                        self.printer.output("--> " + c.str_lhs() + " + (M - b) " + ast[1] + " <= M")
                    c.setb(M)
                    c.setTerm(ast[1], M - b)
                    res.append(c.getAST())
                else:
                    gc = guardedconstraint.GuardedConstraint()
                    gc.setLhs(ast[1])    # y above
                    gc.setRhs(c)
                    res.append(gc.getAST())
        return res

########### end Convex2 class


