#! /usr/bin/env python

import ply.yacc as yacc
import sys

# Get the token map from the lexer. This is required.
from lex3 import tokens


# Variables' bounds management
# Needed to store lower and upper bounds for each declared variable
# Vars is a dictionary: {('x1': [lower_x1, upper_x1]), ..., ('xn': [lower_xn, upper_xn])}
Vars = {}

def addVarBound(var, lower, upper):
    global Vars
    Vars[var] = [lower, upper]

def isVar(str):
    " a variable can be also primed "
    global Vars
    str2 = str.partition("\'")[0]
    return str2 in Vars

def getBound(var):
    if var in Vars:
        return Vars[var]
    else:
        return None

def getBounds():
    " a constraint has a field called 'bounds' with the same type as Vars "
    return Vars
###

# Precedence declaration
precedence = (
    ('nonassoc', 'LEQ', 'GEQ', 'EQ'),  # Nonassociative operators
    ('left', 'TIMES', 'DIVIDE'),
    ('left', 'PLUS', 'MINUS'),
    ('right', 'UMINUS'),
    ('left', 'OR'),
    ('left', 'AND'),
)

# Grammars' productions are defined through functions definition
# See grammar.txt for a detailed explanation of the grammar.
#
# Grammar
# 
# Rule 1     main -> bounds
# Rule 2     main -> formula
# Rule 3     main -> empty
# Rule 4     empty -> <empty>
# Rule 5     bounds -> ID ASSIGN expr TO expr
# Rule 6     formula -> LBRACE formula RBRACE
# Rule 7     formula -> formula AND formula
# Rule 8     formula -> formula OR formula
# Rule 9     formula -> atom
# Rule 10    atom -> equation
# Rule 11    equation -> polynomial LEQ polynomial
# Rule 12    equation -> polynomial GEQ polynomial
# Rule 13    equation -> polynomial EQ polynomial
# Rule 14    polynomial -> monomial
# Rule 15    polynomial -> polynomial PLUS monomial
# Rule 16    polynomial -> polynomial MINUS monomial
# Rule 17    monomial -> term ID
# Rule 18    monomial -> term
# Rule 19    monomial -> ID
# Rule 20    monomial -> MINUS ID
# Rule 21    monomial -> PLUS ID
# Rule 22    expr -> expr PLUS expr
# Rule 23    expr -> expr MINUS expr
# Rule 24    expr -> term
# Rule 25    term -> term TIMES term
# Rule 26    term -> term DIVIDE term
# Rule 27    term -> factor
# Rule 28    factor -> number
# Rule 29    factor -> LPAREN expr RPAREN
# Rule 30    number -> MINUS abs_number
# Rule 31    number -> PLUS abs_number
# Rule 32    number -> abs_number
# Rule 33    abs_number -> NUMBER
# Rule 34    abs_number -> INFTY
# Rule 35    abs_number -> REALNUMBER

def p_main_bounds(p):
    'main : bounds'
    p[0] = []

def p_main_formula(p):
    'main : formula'
    p[0] = p[1]

def p_main_empty(p):
    'main : empty'

def p_empty(p):
    'empty :'
    p[0] = []
    pass

def p_bounds(p):
    'bounds : ID ASSIGN expr TO expr'
    p[0] = []
    if p[3] > p[5]:
        print "Error in '" + p[1] + "' bounds: " + str(p[3]) + " > " + str(p[5]) + "."
    addVarBound(p[1], p[3], p[5])

#########

def p_formula_PAREN(p):
    'formula : LBRACE formula RBRACE'
    p[0] = p[2]

def p_formula_AND_formula(p):
    'formula : formula AND formula'
    p[0] = ['AND', p[1], p[3]]

def p_formula_OR_formula(p):
    'formula : formula OR formula'
    p[0] = ['OR', p[1], p[3]]

def p_formula_atom(p):
    'formula : atom'
    p[0] = p[1]

def p_atom_eq(p):
    'atom : equation'
    p[0] = p[1]

#############

def p_equation_LEQ(p):
    'equation : polynomial LEQ polynomial'
    p[0] = ['LEQ', p[1], p[3]]

def p_equation_GEQ(p):
    'equation : polynomial GEQ polynomial'
    p[0] = ['GEQ', p[1], p[3]]

def p_equation_EQ(p):
    'equation : polynomial EQ polynomial'
    p[0] = ['EQ', p[1], p[3]]

def p_polynomial_monomial(p):
    'polynomial : monomial'
    p[0] = p[1]

def p_polynomial_polynomial_PLUS_monomial(p):
    'polynomial : polynomial PLUS monomial'
    p[0] = ['PLUS', p[1], p[3]]

def p_polynomial_polynomial_MINUS_monomial(p):
    'polynomial : polynomial MINUS monomial'
    p[0] = ['MINUS', p[1], p[3]]

def p_monomial_term_ID(p):
    'monomial : term ID'
    if isVar(p[2]):
        p[0] = ['MONOMIAL', {p[2]: p[1]}]
    else:
        print "Syntax error: '" + p[2] + "' not declared!"
        sys.exit()

def p_monomial_term(p):
    'monomial : term'
    p[0] = ['CONST', p[1]]

def p_monomial_ID(p):
    'monomial : ID'
    if isVar(p[1]):
        p[0] = ['MONOMIAL', {p[1]: 1}]
    else:
        print "Syntax error: '" + p[1] + "' not declared!"
        sys.exit()

def p_monomial_MINUS_ID(p):
    'monomial : MINUS ID %prec MINUS'
    if isVar(p[2]):
        p[0] = ['MONOMIAL', {p[2]: -1}]
    else:
        print "Syntax error: '" + p[2] + "' not declared!"
        sys.exit()

def p_monomial_PLUS_ID(p):
    'monomial : PLUS ID %prec MINUS'
    if isVar(p[2]):
        p[0] = ['MONOMIAL', {p[2]: 1}]
    else:
        print "Syntax error: '" + p[2] + "' not declared!"
        sys.exit()

def p_expr_PLUS_expr(p):
    'expr : expr PLUS expr'
    p[0] = p[1] + p[3]

def p_expr_MINUS_expr(p):
    'expr : expr MINUS expr'
    p[0] = p[1] - p[3]

def p_expr_term(p):
    'expr : term'
    p[0] = p[1]

def p_term_TIMES_term(p):
    'term : term TIMES term'
    p[0] = float(p[1] * p[3])

def p_term_DIVIDE_term(p):
    'term : term DIVIDE term'
    p[0] = float(p[1] / p[3])

def p_term_factor(p):
    'term : factor'
    p[0] = p[1]

def p_factor_number(p):
    'factor : number'
    p[0] = p[1]

def p_factor_PARENS_expr(p):
    'factor : LPAREN expr RPAREN'
    p[0] = p[2]

def p_number_MINUS_abs_number(p):
    'number : MINUS abs_number %prec UMINUS'
    p[0] = -p[2]

def p_number_PLUS_abs_number(p):
    'number : PLUS abs_number %prec UMINUS'
    p[0] = p[2]

def p_number_abs_number(p):
    'number : abs_number'
    p[0] = p[1]

def p_abs_number_NUMBER(p):
    'abs_number : NUMBER'
    p[0] = int(p[1])

def p_abs_number_INFTY(p):
    'abs_number : INFTY'
    p[0] = float('infinity')

def p_abs_number_REALNUMBER(p):
    'abs_number : REALNUMBER'
    p[0] = float(p[1])


# Error rule for syntax errors
def p_error(p):
    print "Syntax error in input! Unexpected token " + str(p.value) + " (" + p.type + ")"
    sys.exit()

# Build the parser
parser = yacc.yacc()


# just for debug
if (__name__=="__main__"):
    for line in sys.stdin:
#        result = parser.parse(line, debug=10)
        result = parser.parse(line)
        if result != [] and result != None:
            print line.strip(), ":",
            print result
    print Vars
